import logging
import re
from typing import Optional
from dependency_injector.wiring import inject, Provide
from flask import Blueprint, request, url_for, jsonify, abort
from pydantic import BaseModel, Field, ValidationError, validator
from udata_search_service.container import Container
from udata_search_service.config import Config
from udata_search_service.infrastructure.services import DatasetService, OrganizationService, ReuseService, DataserviceService
from udata_search_service.infrastructure.search_clients import ElasticClient
from udata_search_service.infrastructure.consumers import DatasetConsumer, ReuseConsumer, OrganizationConsumer, DataserviceConsumer
from udata_search_service.infrastructure.migrate import set_alias as set_alias_func
from udata_search_service.presentation.utils import is_list_type


bp = Blueprint('api', __name__, url_prefix='/api/1')


class OrganizationArgs(BaseModel):
    q: Optional[str] = None
    page: Optional[int] = 1
    page_size: Optional[int] = 20
    sort: Optional[str] = None
    badge: Optional[str] = None

    @validator('sort')
    def sort_validate(cls, value):
        sorts = [
            'created', 'reuses', 'datasets', 'followers', 'views'
        ]
        choices = sorts + ['-' + k for k in sorts]
        if value not in choices:
            raise ValueError('Sort parameter is not in the sorts available choices.')
        return value


class DatasetArgs(BaseModel):
    q: Optional[str] = None
    page: Optional[int] = 1
    page_size: Optional[int] = 20
    sort: Optional[str] = None
    tag: Optional[list[str]] = None
    badge: Optional[str] = None
    organization: Optional[str] = None
    organization_badge: Optional[str] = None
    owner: Optional[str] = None
    license: Optional[str] = None
    geozone: Optional[str] = None
    granularity: Optional[str] = None
    format: Optional[str] = None
    temporal_coverage: Optional[str] = None
    featured: Optional[str] = None
    schema_: Optional[str] = Field(None, alias="schema")
    topic: Optional[str] = None

    @validator('temporal_coverage')
    def temporal_coverage_format(cls, value):
        pattern = re.compile("^([0-9]{4}-[0-9]{2}-[0-9]{2}-[0-9]{4}-[0-9]{2}-[0-9]{2})$")
        if not pattern.match(value):
            raise ValueError('Temporal coverage does not match the right pattern.')
        return value

    @validator('sort')
    def sort_validate(cls, value):
        sorts = [
            'created', 'last_update', 'reuses', 'followers', 'views'
        ]
        choices = sorts + ['-' + k for k in sorts]
        if value not in choices:
            raise ValueError('Temporal coverage does not match the right pattern.')
        return value

    @classmethod
    def from_request_args(cls, request_args) -> 'DatasetArgs':
        def get_list_args() -> dict:
            return {
                key: value
                for key, value in request_args.to_dict(flat=False).items()
                if key in cls.__fields__
                and is_list_type(cls.__fields__[key].annotation)
            }

        return cls(
            **{
                **request_args.to_dict(),
                **get_list_args(),
            }
        )

class ReuseArgs(BaseModel):
    q: Optional[str] = None
    page: Optional[int] = 1
    page_size: Optional[int] = 20
    sort: Optional[str] = None
    tag: Optional[str] = None
    badge: Optional[str] = None
    organization: Optional[str] = None
    organization_badge: Optional[str] = None
    owner: Optional[str] = None
    type: Optional[str] = None
    featured: Optional[str] = None
    topic: Optional[str] = None

    @validator('sort')
    def sort_validate(cls, value):
        sorts = [
            'created', 'datasets', 'followers', 'views'
        ]
        choices = sorts + ['-' + k for k in sorts]
        if value not in choices:
            raise ValueError('Temporal coverage does not match the right pattern.')
        return value


class DataserviceArgs(BaseModel):
    q: Optional[str] = None
    page: Optional[int] = 1
    page_size: Optional[int] = 20
    sort: Optional[str] = None
    tag: Optional[str] = None
    organization: Optional[str] = None
    owner: Optional[str] = None
    is_restricted: Optional[bool] = None

    @validator('sort')
    def sort_validate(cls, value):
        sorts = [
            'created', 'followers', 'views'
        ]
        choices = sorts + ['-' + k for k in sorts]
        if value not in choices:
            raise ValueError('Sort parameter is not in the sorts available choices.')
        return value


def make_response(results, total_pages, results_number, page, page_size, next_url, prev_url):
    return jsonify({
        "data": results,
        "next_page": next_url if page < total_pages else None,
        "page": page,
        "previous_page": prev_url if page > 1 else None,
        "page_size": page_size,
        "total_pages": total_pages,
        "total": results_number
    })


class DatasetToIndex(BaseModel):
    id: str
    title: str
    description: str
    acronym: Optional[str] = None
    url: Optional[str] = None
    tags: Optional[list] = []
    license: Optional[str] = None
    badges: Optional[list] = []
    frequency: Optional[str] = None
    created_at: str
    last_update: str
    organization: Optional[dict] = None
    owner: Optional[str] = None
    views: int
    followers: int
    reuses: int
    resources_count: Optional[int] = 0
    resources: Optional[list] = None
    featured: Optional[int] = 0
    format: Optional[list] = []
    schema_: Optional[list] = Field([], alias="schema")
    extras: Optional[dict] = {}
    harvest: Optional[dict] = {}
    temporal_coverage_start: Optional[str] = None
    temporal_coverage_end: Optional[str] = None
    geozones: Optional[list] = []
    granularity: Optional[str] = None
    topics: Optional[list] = []


class RequestDatasetIndex(BaseModel):
    document: DatasetToIndex
    index: Optional[str] = None


@bp.route("/datasets/index", methods=["POST"], endpoint='dataset_index')
@inject
def dataset_index(dataset_service: DatasetService = Provide[Container.dataset_service], search_client: ElasticClient = Provide[Container.search_client]):
    try:
        validated_obj = RequestDatasetIndex(**request.json)
    except ValidationError as e:
        abort(400, e)

    document = DatasetConsumer.load_from_dict(validated_obj.document.dict())
    index_name = f'{Config.UDATA_INSTANCE_NAME}-{validated_obj.index}' if validated_obj.index else None
    if index_name and not search_client.es.indices.exists(index=index_name):
        abort(404, 'Index does not exist')
    dataset_service.feed(document, index_name)
    return jsonify({'data': 'Dataset added to index'})


@bp.route("/datasets/<dataset_id>/unindex", methods=["DELETE"], endpoint='dataset_unindex')
@inject
def dataset_unindex(dataset_id: str, dataset_service: DatasetService = Provide[Container.dataset_service]):
    result = dataset_service.delete_one(dataset_id)
    if result:
        return jsonify({'data': f'Dataset {result} removed from index'})
    abort(404, 'reuse not found')


@bp.route("/datasets/", methods=["GET"], endpoint='dataset_search')
@inject
def datasets_search(dataset_service: DatasetService = Provide[Container.dataset_service]):
    try:
        request_args = DatasetArgs.from_request_args(request.args)
    except ValidationError as e:
        abort(400, e)

    results, results_number, total_pages = dataset_service.search(request_args.dict())

    next_url = url_for('api.dataset_search', q=request_args.q, page=request_args.page + 1,
                       page_size=request_args.page_size, _external=True)
    prev_url = url_for('api.dataset_search', q=request_args.q, page=request_args.page - 1,
                       page_size=request_args.page_size, _external=True)

    return make_response(results, total_pages, results_number,
                         request_args.page, request_args.page_size, next_url, prev_url)


@bp.route("/datasets/<dataset_id>/", methods=["GET"], endpoint='dataset_get_specific')
@inject
def get_dataset(dataset_id: str, dataset_service: DatasetService = Provide[Container.dataset_service]):
    result = dataset_service.find_one(dataset_id)
    if result:
        return jsonify(result)
    abort(404, 'dataset not found')


@bp.route("/organizations/", methods=["GET"], endpoint='organization_search')
@inject
def organizations_search(organization_service: OrganizationService = Provide[Container.organization_service]):
    try:
        request_args = OrganizationArgs(**request.args)
    except ValidationError as e:
        abort(400, e)

    results, results_number, total_pages = organization_service.search(request_args.dict())

    next_url = url_for('api.organization_search', q=request_args.q, page=request_args.page + 1,
                       page_size=request_args.page_size, _external=True)
    prev_url = url_for('api.organization_search', q=request_args.q, page=request_args.page - 1,
                       page_size=request_args.page_size, _external=True)

    return make_response(results, total_pages, results_number, request_args.page,
                         request_args.page_size, next_url, prev_url)


@bp.route("/organizations/<organization_id>/", methods=["GET"], endpoint='organization_get_specific')
@inject
def get_organization(organization_id: str,
                     organization_service: OrganizationService = Provide[Container.organization_service]):
    result = organization_service.find_one(organization_id)
    if result:
        return jsonify(result)
    abort(404, 'organization not found')


class OrganizationToIndex(BaseModel):
    id: str
    name: str
    description: str
    acronym: Optional[str] = None
    url: Optional[str] = None
    badges: Optional[list] = []
    created_at: str
    orga_sp: int
    datasets: int
    followers: int
    reuses: int
    views: int
    extras: Optional[dict] = {}


class RequestOrganizationIndex(BaseModel):
    document: OrganizationToIndex
    index: Optional[str] = None


@bp.route("/organizations/index", methods=["POST"], endpoint='organization_index')
@inject
def organization_index(organization_service: OrganizationService = Provide[Container.organization_service], search_client: ElasticClient = Provide[Container.search_client]):
    try:
        validated_obj = RequestOrganizationIndex(**request.json)
    except ValidationError as e:
        abort(400, e)

    document = OrganizationConsumer.load_from_dict(validated_obj.document.dict())
    index_name = f'{Config.UDATA_INSTANCE_NAME}-{validated_obj.index}' if validated_obj.index else None
    if index_name and not search_client.es.indices.exists(index=index_name):
        abort(404, 'Index does not exist')
    organization_service.feed(document, index_name)
    return jsonify({'data': 'Organization added to index'})


@bp.route("/organizations/<organization_id>/unindex", methods=["DELETE"], endpoint='organization_unindex')
@inject
def organization_unindex(organization_id: str, organization_service: OrganizationService = Provide[Container.organization_service]):
    result = organization_service.delete_one(organization_id)
    if result:
        return jsonify({'data': f'Organization {result} removed from index'})
    abort(404, 'reuse not found')


@bp.route("/reuses/", methods=["GET"], endpoint='reuse_search')
@inject
def reuses_search(reuse_service: ReuseService = Provide[Container.reuse_service]):
    try:
        request_args = ReuseArgs(**request.args)
    except ValidationError as e:
        abort(400, e)

    results, results_number, total_pages = reuse_service.search(request_args.dict())

    next_url = url_for('api.reuse_search', q=request_args.q, page=request_args.page + 1,
                       page_size=request_args.page_size, _external=True)
    prev_url = url_for('api.reuse_search', q=request_args.q, page=request_args.page - 1,
                       page_size=request_args.page_size, _external=True)

    return make_response(results, total_pages, results_number, request_args.page,
                         request_args.page_size, next_url, prev_url)


@bp.route("/reuses/<reuse_id>/", methods=["GET"], endpoint='reuse_get_specific')
@inject
def get_reuse(reuse_id: str, reuse_service: ReuseService = Provide[Container.reuse_service]):
    result = reuse_service.find_one(reuse_id)
    if result:
        return jsonify(result)
    abort(404, 'reuse not found')


class ReuseToIndex(BaseModel):
    id: str
    title: str
    description: str
    url: Optional[str] = None
    badges: Optional[list] = []
    created_at: str
    archived: Optional[str] = None
    datasets: int
    followers: int
    views: int
    featured: int
    organization: Optional[dict] = {}
    owner: Optional[str] = None
    type: str
    topic: str
    tags: Optional[list] = []
    extras: Optional[dict] = {}


class RequestReuseIndex(BaseModel):
    document: ReuseToIndex
    index: Optional[str] = None


@bp.route("/reuses/index", methods=["POST"], endpoint='reuse_index')
@inject
def reuse_index(reuse_service: ReuseService = Provide[Container.reuse_service], search_client: ElasticClient = Provide[Container.search_client]):
    try:
        validated_obj = RequestReuseIndex(**request.json)
    except ValidationError as e:
        abort(400, e)

    document = ReuseConsumer.load_from_dict(validated_obj.document.dict())
    index_name = f'{Config.UDATA_INSTANCE_NAME}-{validated_obj.index}' if validated_obj.index else None
    if index_name and not search_client.es.indices.exists(index=index_name):
        abort(404, 'Index does not exist')

    reuse_service.feed(document, index_name)
    return jsonify({'data': 'Reuse added to index'})


@bp.route("/reuses/<reuse_id>/unindex", methods=["DELETE"], endpoint='reuse_unindex')
@inject
def reuse_unindex(reuse_id: str, reuse_service: ReuseService = Provide[Container.reuse_service]):
    result = reuse_service.delete_one(reuse_id)
    if result:
        return jsonify({'data': f'Reuse {result} removed from index'})
    abort(404, 'reuse not found')


@bp.route("/dataservices/", methods=["GET"], endpoint='dataservice_search')
@inject
def dataservices_search(dataservice_service: DataserviceService = Provide[Container.dataservice_service]):
    try:
        request_args = DataserviceArgs(**request.args)
    except ValidationError as e:
        abort(400, e)

    results, results_number, total_pages = dataservice_service.search(request_args.dict())

    next_url = url_for('api.dataservice_search', q=request_args.q, page=request_args.page + 1,
                       page_size=request_args.page_size, _external=True)
    prev_url = url_for('api.dataservice_search', q=request_args.q, page=request_args.page - 1,
                       page_size=request_args.page_size, _external=True)

    return make_response(results, total_pages, results_number, request_args.page,
                         request_args.page_size, next_url, prev_url)


@bp.route("/dataservices/<dataservice_id>/", methods=["GET"], endpoint='dataservice_get_specific')
@inject
def get_dataservice(dataservice_id: str, dataservice_service: DataserviceService = Provide[Container.dataservice_service]):
    result = dataservice_service.find_one(dataservice_id)
    if result:
        return jsonify(result)
    abort(404, 'dataservice not found')


class DataserviceToIndex(BaseModel):
    id: str
    title: str
    description: str
    created_at: str
    views: int = 0
    followers: int = 0
    organization: Optional[dict] = {}
    owner: Optional[str] = None
    tags: Optional[list] = []
    extras: Optional[dict] = {}
    is_restricted: Optional[bool] = None


class RequestDataserviceIndex(BaseModel):
    document: DataserviceToIndex
    index: Optional[str] = None


@bp.route("/dataservices/index", methods=["POST"], endpoint='dataservice_index')
@inject
def dataservice_index(dataservice_service: DataserviceService = Provide[Container.dataservice_service], search_client: ElasticClient = Provide[Container.search_client]):
    try:
        validated_obj = RequestDataserviceIndex(**request.json)
    except ValidationError as e:
        abort(400, e)

    document = DataserviceConsumer.load_from_dict(validated_obj.document.dict())
    index_name = f'{Config.UDATA_INSTANCE_NAME}-{validated_obj.index}' if validated_obj.index else None
    if index_name and not search_client.es.indices.exists(index=index_name):
        abort(404, 'Index does not exist')

    dataservice_service.feed(document, index_name)
    return jsonify({'data': 'Dataservice added to index'})


@bp.route("/dataservices/<dataservice_id>/unindex", methods=["DELETE"], endpoint='dataservice_unindex')
@inject
def dataservice_unindex(dataservice_id: str, dataservice_service: DataserviceService = Provide[Container.dataservice_service]):
    result = dataservice_service.delete_one(dataservice_id)
    if result:
        return jsonify({'data': f'Dataservice {result} removed from index'})
    abort(404, 'dataservice not found')


@bp.route("/create-index", methods=["POST"], endpoint='create_index')
@inject
def create_index(search_client: ElasticClient = Provide[Container.search_client]):
    try:
        index_name = request.json['index']
    except KeyError:
        abort(400, 'Missing index in payload')

    index_name = f'{Config.UDATA_INSTANCE_NAME}-{index_name}'
    # Initiliaze index matching template pattern
    if not search_client.es.indices.exists(index=index_name):
        logging.info(f'Initializing new index {index_name} for reindexation')
        search_client.es.indices.create(index=index_name)
        # Check whether template with analyzer was correctly assigned
        if 'analysis' not in search_client.es.indices.get_settings(index=index_name)[index_name]['settings']['index']:
            logging.error(f'Analyzer was not set using templates when initializing {index_name}')
        return jsonify({'data': f'Index {index_name} created'})
    return jsonify({'data': f'Index {index_name} already exists'})


class RequestSetAlias(BaseModel):
    index_suffix_name: str
    indices: Optional[list] = []


@bp.route("/set-index-alias", methods=["POST"], endpoint='set_index_alias')
@inject
def set_index_alias(search_client: ElasticClient = Provide[Container.search_client]):
    try:
        validated_obj = RequestSetAlias(**request.json).dict()
    except ValidationError as e:
        abort(400, e)

    indices = [index.lower().rstrip('s') for index in (validated_obj['indices'] or [])]
    set_alias_func(search_client, validated_obj['index_suffix_name'], indices)
    return jsonify({'data': 'Alias set'})



from typing import Tuple, Optional, List
from udata_search_service.domain.entities import Dataset, Organization, Reuse, Dataservice
from udata_search_service.infrastructure.search_clients import ElasticClient


class OrganizationService:

    def __init__(self, search_client: ElasticClient):
        self.search_client = search_client

    def feed(self, organization: Organization, index: str = None) -> None:
        self.search_client.index_organization(organization, index)

    def search(self, filters: dict) -> Tuple[List[Organization], int, int]:
        page = filters.pop('page')
        page_size = filters.pop('page_size')
        search_text = filters.pop('q')
        sort = self.format_sort(filters.pop('sort', None))

        if page > 1:
            offset = page_size * (page - 1)
        else:
            offset = 0

        self.format_filters(filters)

        results_number, search_results = self.search_client.query_organizations(search_text, offset, page_size, filters, sort)
        results = [Organization.load_from_dict(hit) for hit in search_results]
        total_pages = round(results_number / page_size) or 1
        return results, results_number, total_pages

    def find_one(self, organization_id: str) -> Optional[Organization]:
        try:
            return Organization.load_from_dict(self.search_client.find_one_organization(organization_id))
        except TypeError:
            return None

    def delete_one(self, organization_id: str) -> Optional[str]:
        return self.search_client.delete_one_organization(organization_id)

    @staticmethod
    def format_filters(filters):
        filtered = {k: v for k, v in filters.items() if v is not None}
        filters.clear()
        filters.update(filtered)

    @staticmethod
    def format_sort(sort):
        if sort and 'created' in sort:
            sort = sort.replace('created', 'created_at')
        return sort


class DatasetService:

    def __init__(self, search_client: ElasticClient):
        self.search_client = search_client

    def feed(self, dataset: Dataset, index: str = None) -> None:
        self.search_client.index_dataset(dataset, index)

    def search(self, filters: dict) -> Tuple[List[Dataset], int, int]:
        page = filters.pop('page')
        page_size = filters.pop('page_size')
        search_text = filters.pop('q')
        sort = self.format_sort(filters.pop('sort', None))

        if page > 1:
            offset = page_size * (page - 1)
        else:
            offset = 0

        self.format_filters(filters)

        results_number, search_results = self.search_client.query_datasets(search_text, offset, page_size, filters, sort)
        results = [Dataset.load_from_dict(hit) for hit in search_results]
        total_pages = round(results_number / page_size) or 1
        return results, results_number, total_pages

    def find_one(self, dataset_id: str) -> Optional[Dataset]:
        try:
            return Dataset.load_from_dict(self.search_client.find_one_dataset(dataset_id))
        except TypeError:
            return None

    def delete_one(self, dataset_id: str) -> Optional[str]:
        return self.search_client.delete_one_dataset(dataset_id)

    @staticmethod
    def format_filters(filters):
        '''
        Format search filters params to match the actual fields in ElasticSearch.
        For example udata search params uses singular ?tag=<mytag>, even though
        the field is plural since it's a list of tags.
        '''
        if filters['temporal_coverage']:
            parts = filters.pop('temporal_coverage')
            filters['temporal_coverage_start'] = parts[:10]
            filters['temporal_coverage_end'] = parts[11:]
        if filters['tag']:
            filters['tags'] = filters.pop('tag')
        if filters['topic']:
            filters['topics'] = filters.pop('topic')
        if filters['geozone']:
            filters['geozones'] = filters.pop('geozone')
        if filters['schema_']:
            filters['schema'] = filters.pop('schema_')
        if filters['organization_badge']:
            filters['organization_badges'] = filters.pop('organization_badge')
        filtered = {k: v for k, v in filters.items() if v is not None}
        filters.clear()
        filters.update(filtered)

    @staticmethod
    def format_sort(sort):
        if sort is not None and 'created' in sort:
            sort = sort.replace('created', 'created_at')
        return sort


class ReuseService:

    def __init__(self, search_client: ElasticClient):
        self.search_client = search_client

    def feed(self, reuse: Reuse, index: str = None) -> None:
        self.search_client.index_reuse(reuse, index)

    def search(self, filters: dict) -> Tuple[List[Reuse], int, int]:
        page = filters.pop('page')
        page_size = filters.pop('page_size')
        search_text = filters.pop('q')
        sort = self.format_sort(filters.pop('sort', None))

        if page > 1:
            offset = page_size * (page - 1)
        else:
            offset = 0

        self.format_filters(filters)

        results_number, search_results = self.search_client.query_reuses(search_text, offset, page_size, filters, sort)
        results = [Reuse.load_from_dict(hit) for hit in search_results]
        total_pages = round(results_number / page_size) or 1
        return results, results_number, total_pages

    def find_one(self, reuse_id: str) -> Optional[Reuse]:
        try:
            return Reuse.load_from_dict(self.search_client.find_one_reuse(reuse_id))
        except TypeError:
            return None

    def delete_one(self, reuse_id: str) -> Optional[str]:
        return self.search_client.delete_one_reuse(reuse_id)

    @staticmethod
    def format_filters(filters):
        if filters['tag']:
            filters['tags'] = filters.pop('tag')
        if filters['organization_badge']:
            filters['organization_badges'] = filters.pop('organization_badge')
        filtered = {k: v for k, v in filters.items() if v is not None}
        filters.clear()
        filters.update(filtered)

    @staticmethod
    def format_sort(sort):
        if sort is not None and 'created' in sort:
            sort = sort.replace('created', 'created_at')
        return sort


class DataserviceService:

    def __init__(self, search_client: ElasticClient):
        self.search_client = search_client

    def feed(self, dataservice: Dataservice, index: str = None) -> None:
        self.search_client.index_dataservice(dataservice, index)

    def search(self, filters: dict) -> Tuple[List[Dataservice], int, int]:
        page = filters.pop('page')
        page_size = filters.pop('page_size')
        search_text = filters.pop('q')
        sort = self.format_sort(filters.pop('sort', None))

        if page > 1:
            offset = page_size * (page - 1)
        else:
            offset = 0

        self.format_filters(filters)

        results_number, search_results = self.search_client.query_dataservices(search_text, offset, page_size, filters, sort)
        results = [Dataservice.load_from_dict(hit) for hit in search_results]
        total_pages = round(results_number / page_size) or 1
        return results, results_number, total_pages

    def find_one(self, dataservice_id: str) -> Optional[Dataservice]:
        try:
            return Dataservice.load_from_dict(self.search_client.find_one_dataservice(dataservice_id))
        except TypeError:
            return None

    def delete_one(self, dataservice_id: str) -> Optional[str]:
        return self.search_client.delete_one_dataservice(dataservice_id)

    @staticmethod
    def format_filters(filters):
        if filters['tag']:
            filters['tags'] = filters.pop('tag')
        filtered = {k: v for k, v in filters.items() if v is not None}
        filters.clear()
        filters.update(filtered)

    @staticmethod
    def format_sort(sort):
        if sort is not None and 'created' in sort:
            sort = sort.replace('created', 'created_at')
        return sort



import os


class Config:
    SECRET_KEY = os.environ.get('SECRET_KEY') or 'you-will-never-guess'

    ELASTICSEARCH_URL = os.environ.get('ELASTICSEARCH_URL') or 'http://localhost:9200'

    UDATA_INSTANCE_NAME = os.environ.get('UDATA_INSTANCE_NAME') or 'udata'

    SEARCH_SYNONYMS = [
        "AMD, administrateur ministériel des données, AMDAC",
        "lolf, loi de finance",
        "waldec, RNA, répertoire national des associations",
        "ovq, baromètre des résultats",
        "contour, découpage",
        "rp, recensement de la population"
    ]


class Testing(Config):
    TESTING = True
    ELASTICSEARCH_URL = os.environ.get('ELASTICSEARCH_URL_TEST') or 'http://localhost:9201'



from faker import Faker
import pytest
from udata_search_service.app import create_app
from udata_search_service.config import Testing
from udata_search_service.infrastructure.search_clients import SearchableDataset, SearchableReuse, SearchableOrganization, ElasticClient


@pytest.fixture
def app():
    app = create_app(Testing)
    yield app
    app.container.unwire()


@pytest.fixture
def client(app):
    with app.test_client() as client:
        return client


@pytest.fixture
def search_client(app):
    return ElasticClient(url=app.config['ELASTICSEARCH_URL'])


@pytest.fixture(autouse=True)
def db(search_client):

    search_client.clean_indices()

    yield

    search_client.es.indices.delete(index="*")


@pytest.fixture
def faker():
    return Faker()



from math import log
import ssl
import sys
from tempfile import _TemporaryFileWrapper
from urllib.request import urlopen

from bs4 import BeautifulSoup
from markdown import markdown

IS_TTY = sys.__stdin__.isatty()


def download_catalog(url: str, fd: _TemporaryFileWrapper) -> None:
    ssl._create_default_https_context = ssl._create_unverified_context
    response = urlopen(url)
    while True:
        chunk: bytes = response.read(1024)
        if not chunk:
            break
        fd.write(chunk)


def get_concat_title_org(title: str, acronym: str, organization_name: str) -> str:
    concat = title
    if acronym:
        concat += ' ' + acronym
    if organization_name:
        concat += ' ' + organization_name
    return concat


def log2p(value):
    '''
    Add 2 to the field value and take the common logarithm.
    It makes sure that the result is > 0, needed for function score
    Using multiply boost mode
    '''
    if not value:
        value = 0
    return log(value + 2)


def mdstrip(value):
    '''
    Truncate and strip tags from a markdown source

    The markdown source is truncated at the excerpt if present and
    smaller than the required length. Then, all html tags are stripped.
    '''
    if not value:
        return ''
    rendered = markdown(value, wrap=False)
    text = ''.join(BeautifulSoup(rendered, 'html.parser').findAll(text=True))
    return text



import datetime
import time
from flask import url_for

from udata_search_service.domain.factories import DatasetFactory, OrganizationFactory, ReuseFactory


def test_api_dataset_index_unindex(app, client, faker):
    dataset = {
        'id': faker.md5(),
        'title': faker.sentence(),
        'description': faker.text(),
        'acronym': faker.company_suffix(),
        'url': faker.url(),
        'created_at': faker.past_datetime().isoformat(),
        'last_update': faker.past_datetime().isoformat(),
        'views': faker.random_int(),
        'followers': faker.random_int(),
        'reuses': faker.random_int(),
        'featured': faker.random_int(min=0, max=1),
        'resources_count': faker.random_int(min=1, max=15),
        'resources': [
            {"id": faker.md5(), "title": faker.word()}
            for _ in range(5)
        ],
        'organization': {
            'id': faker.md5(),
            'name': faker.company(),
            'public_service': faker.random_int(min=0, max=1),
            'followers': faker.random_int(),
            'badges': [faker.word()]
        },
        'format': ['pdf'],
        'frequency': 'unknown',
        'badges': [],
        'tags': [faker.word()],
        'license': faker.word(),
        'temporal_coverage_start': faker.past_datetime().isoformat(),
        'temporal_coverage_end': faker.past_datetime().isoformat(),
        'granularity': faker.word(),
        'geozones': [{'id': faker.word(), 'name': faker.word(), 'keys': [faker.random_int()]},
                     {'id': faker.word()}],
        'owner': None,
        'extras': {},
        'harvest': {},
        'schema': [faker.word(), faker.word()]
    }

    query = {
        'document': dataset,
        'index': None
    }

    index_resp = client.post(url_for('api.dataset_index'), json={'document': dataset, 'index': 'random-non-existing-index'})
    assert index_resp.status_code == 404

    index_resp = client.post(url_for('api.dataset_index'), json=query)
    assert index_resp.status_code == 200

    time.sleep(2)

    dataset_resp = client.get(url_for('api.dataset_get_specific', dataset_id=dataset['id']))
    assert dataset_resp.status_code == 200
    assert dataset_resp.json['title'] == dataset['title']

    dataset_search_resp = client.get(url_for('api.dataset_search'))
    assert len(dataset_search_resp.json['data']) == 1
    assert dataset_search_resp.json['data'][0]['title'] == dataset['title']
    assert dataset_search_resp.json['next_page'] is None
    assert dataset_search_resp.json['page'] == 1
    assert dataset_search_resp.json['previous_page'] is None
    assert dataset_search_resp.json['page_size'] == 20
    assert dataset_search_resp.json['total_pages'] == 1
    assert dataset_search_resp.json['total'] == 1

    deletion_resp = client.delete(url_for('api.dataset_unindex', dataset_id=dataset['id']))
    assert deletion_resp.status_code == 200

    time.sleep(2)

    dataset_get_after_delete_resp = client.get(url_for('api.dataset_get_specific', dataset_id=dataset['id']))
    assert dataset_get_after_delete_resp.status_code == 404

    dataset_search_after_delete_resp = client.get(url_for('api.dataset_search'))
    assert len(dataset_search_after_delete_resp.json['data']) == 0
    assert dataset_search_after_delete_resp.json['next_page'] is None
    assert dataset_search_after_delete_resp.json['page'] == 1
    assert dataset_search_after_delete_resp.json['previous_page'] is None
    assert dataset_search_after_delete_resp.json['page_size'] == 20
    assert dataset_search_after_delete_resp.json['total_pages'] == 1
    assert dataset_search_after_delete_resp.json['total'] == 0


def test_api_dataset_index_on_another_index(app, client, search_client, faker):
    now = datetime.datetime.utcnow().strftime('%Y-%m-%d-%H-%M')
    index_name = f"test-dataset-{now}"
    if not search_client.es.indices.exists(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}"):
        search_client.es.indices.create(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}")

    dataset = {
        'id': faker.md5(),
        'title': faker.sentence(),
        'description': faker.text(),
        'acronym': faker.company_suffix(),
        'url': faker.url(),
        'created_at': faker.past_datetime().isoformat(),
        'last_update': faker.past_datetime().isoformat(),
        'views': faker.random_int(),
        'followers': faker.random_int(),
        'reuses': faker.random_int(),
        'featured': faker.random_int(min=0, max=1),
        'resources_count': faker.random_int(min=1, max=15),
        'resources': [
            {"id": faker.md5(), "title": faker.word()}
            for _ in range(5)
        ],
        'organization': {
            'id': faker.md5(),
            'name': faker.company(),
            'public_service': faker.random_int(min=0, max=1),
            'followers': faker.random_int(),
            'badges': [faker.word()]
        },
        'format': ['pdf'],
        'frequency': 'unknown',
        'badges': [],
        'tags': [faker.word()],
        'license': faker.word(),
        'temporal_coverage_start': faker.past_datetime().isoformat(),
        'temporal_coverage_end': faker.past_datetime().isoformat(),
        'granularity': faker.word(),
        'geozones': [{'id': faker.word(), 'name': faker.word(), 'keys': [faker.random_int()]},
                     {'id': faker.word()}],
        'owner': None,
        'extras': {},
        'harvest': {},
        'schema': [faker.word(), faker.word()]
    }

    query = {
        'document': dataset,
        'index': index_name
    }

    index_resp = client.post(url_for('api.dataset_index'), json=query)
    assert index_resp.status_code == 200

    time.sleep(2)

    resp = search_client.es.get(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}", id=dataset['id'])
    assert resp['_source']['title'] == dataset['title']


def test_api_org_index_unindex(app, client, faker):
    org = {
        'id': faker.md5(),
        'name': faker.company(),
        'description': faker.text(),
        'url': faker.url(),
        'created_at': faker.past_datetime().isoformat(),
        'views': faker.random_int(),
        'orga_sp': faker.random_int(),
        'followers': faker.random_int(),
        'datasets': faker.random_int(),
        'reuses': faker.random_int(),
        'badges': [],
        'extras': {},
    }

    query = {
        'document': org,
        'index': None
    }

    index_resp = client.post(url_for('api.organization_index'), json={'document': org, 'index': 'random-non-existing-index'})
    assert index_resp.status_code == 404

    index_resp = client.post(url_for('api.organization_index'), json=query)
    assert index_resp.status_code == 200

    time.sleep(2)

    organization_resp = client.get(url_for('api.organization_get_specific', organization_id=org['id']))
    assert organization_resp.status_code == 200
    assert organization_resp.json['name'] == org['name']

    organization_search_resp = client.get(url_for('api.organization_search'))
    assert len(organization_search_resp.json['data']) == 1
    assert organization_search_resp.json['data'][0]['name'] == org['name']
    assert organization_search_resp.json['next_page'] is None
    assert organization_search_resp.json['page'] == 1
    assert organization_search_resp.json['previous_page'] is None
    assert organization_search_resp.json['page_size'] == 20
    assert organization_search_resp.json['total_pages'] == 1
    assert organization_search_resp.json['total'] == 1

    deletion_resp = client.delete(url_for('api.organization_unindex', organization_id=org['id']))
    assert deletion_resp.status_code == 200

    time.sleep(2)

    organization_get_after_delete_resp = client.get(url_for('api.organization_get_specific', organization_id=org['id']))
    assert organization_get_after_delete_resp.status_code == 404

    organization_search_after_delete_resp = client.get(url_for('api.organization_search'))
    assert len(organization_search_after_delete_resp.json['data']) == 0
    assert organization_search_after_delete_resp.json['next_page'] is None
    assert organization_search_after_delete_resp.json['page'] == 1
    assert organization_search_after_delete_resp.json['previous_page'] is None
    assert organization_search_after_delete_resp.json['page_size'] == 20
    assert organization_search_after_delete_resp.json['total_pages'] == 1
    assert organization_search_after_delete_resp.json['total'] == 0


def test_api_org_index_on_another_index(app, client, search_client, faker):
    now = datetime.datetime.utcnow().strftime('%Y-%m-%d-%H-%M')
    index_name = f"test-organization-{now}"
    if not search_client.es.indices.exists(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}"):
        search_client.es.indices.create(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}")

    org = {
        'id': faker.md5(),
        'name': faker.company(),
        'description': faker.text(),
        'url': faker.url(),
        'created_at': faker.past_datetime().isoformat(),
        'views': faker.random_int(),
        'orga_sp': faker.random_int(),
        'followers': faker.random_int(),
        'datasets': faker.random_int(),
        'reuses': faker.random_int(),
        'badges': [],
        'extras': {},
    }

    query = {
        'document': org,
        'index': index_name
    }

    index_resp = client.post(url_for('api.organization_index'), json=query)
    assert index_resp.status_code == 200

    time.sleep(2)

    resp = search_client.es.get(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}", id=org['id'])
    assert resp['_source']['name'] == org['name']


def test_api_reuse_index_unindex(app, client, faker):
    reuse = {
        'id': faker.md5(),
        'title': faker.sentence(),
        'description': faker.text(),
        'url': faker.url(),
        'badges': [],
        'created_at': faker.past_datetime().isoformat(),
        'datasets': faker.random_int(),
        'views': faker.random_int(),
        'followers': faker.random_int(),
        'featured': faker.random_int(min=0, max=1),
        'organization': {
            'id': faker.md5(),
            'name': faker.company(),
            'public_service': faker.random_int(min=0, max=1),
            'followers': faker.random_int(),
            'badges': [faker.word()]
        },
        'tags': [],
        'owner': None,
        'extras': {},
        'type': faker.word(),
        'topic': faker.word()
    }

    query = {
        'document': reuse,
        'index': None
    }

    index_resp = client.post(url_for('api.reuse_index'), json={'document': reuse, 'index': 'random-non-existing-index'})
    assert index_resp.status_code == 404

    index_resp = client.post(url_for('api.reuse_index'), json=query)
    assert index_resp.status_code == 200

    time.sleep(2)

    reuse_resp = client.get(url_for('api.reuse_get_specific', reuse_id=reuse['id']))
    assert reuse_resp.status_code == 200
    assert reuse_resp.json['title'] == reuse['title']

    reuse_search_resp = client.get(url_for('api.reuse_search'))
    assert len(reuse_search_resp.json['data']) == 1
    assert reuse_search_resp.json['data'][0]['title'] == reuse['title']
    assert reuse_search_resp.json['next_page'] is None
    assert reuse_search_resp.json['page'] == 1
    assert reuse_search_resp.json['previous_page'] is None
    assert reuse_search_resp.json['page_size'] == 20
    assert reuse_search_resp.json['total_pages'] == 1
    assert reuse_search_resp.json['total'] == 1

    deletion_resp = client.delete(url_for('api.reuse_unindex', reuse_id=reuse['id']))
    assert deletion_resp.status_code == 200

    time.sleep(2)

    reuse_get_after_delete_resp = client.get(url_for('api.reuse_get_specific', reuse_id=reuse['id']))
    assert reuse_get_after_delete_resp.status_code == 404

    reuse_search_after_delete_resp = client.get(url_for('api.reuse_search'))
    assert len(reuse_search_after_delete_resp.json['data']) == 0
    assert reuse_search_after_delete_resp.json['next_page'] is None
    assert reuse_search_after_delete_resp.json['page'] == 1
    assert reuse_search_after_delete_resp.json['previous_page'] is None
    assert reuse_search_after_delete_resp.json['page_size'] == 20
    assert reuse_search_after_delete_resp.json['total_pages'] == 1
    assert reuse_search_after_delete_resp.json['total'] == 0


def test_api_reuse_index_on_another_index(app, client, search_client, faker):
    now = datetime.datetime.utcnow().strftime('%Y-%m-%d-%H-%M')
    index_name = f"test-reuse-{now}"
    if not search_client.es.indices.exists(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}"):
        search_client.es.indices.create(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}")

    reuse = {
        'id': faker.md5(),
        'title': faker.sentence(),
        'description': faker.text(),
        'url': faker.url(),
        'badges': [],
        'created_at': faker.past_datetime().isoformat(),
        'datasets': faker.random_int(),
        'views': faker.random_int(),
        'followers': faker.random_int(),
        'featured': faker.random_int(min=0, max=1),
        'organization': {
            'id': faker.md5(),
            'name': faker.company(),
            'public_service': faker.random_int(min=0, max=1),
            'followers': faker.random_int(),
            'badges': [faker.word()]
        },
        'tags': [],
        'owner': None,
        'extras': {},
        'type': faker.word(),
        'topic': faker.word()
    }

    query = {
        'document': reuse,
        'index': index_name
    }

    index_resp = client.post(url_for('api.reuse_index'), json=query)
    assert index_resp.status_code == 200

    time.sleep(2)

    resp = search_client.es.get(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}", id=reuse['id'])
    assert resp['_source']['title'] == reuse['title']


def test_api_dataservice_index_unindex(app, client, faker):
    dataservice = {
        "id": faker.md5(),
        "title": faker.sentence(),
        "description": faker.text(),
        "created_at": faker.past_datetime().isoformat(),
        "followers": faker.random_int(),
        "views": faker.random_int(),
        'organization': {
            'id': faker.md5(),
            'name': faker.company(),
            'public_service': faker.random_int(min=0, max=1),
            'followers': faker.random_int()
        },
        "owner": None,
        "tags": [],
        "is_restricted": False,
        "extras":
        {
            "availability_url": "",
            "is_franceconnect": False,
            "public_cible":
            []
        }
    }

    query = {
        'document': dataservice,
        'index': None
    }

    index_resp = client.post(url_for('api.dataservice_index'), json={'document': dataservice, 'index': 'random-non-existing-index'})
    assert index_resp.status_code == 404

    index_resp = client.post(url_for('api.dataservice_index'), json=query)
    assert index_resp.status_code == 200

    time.sleep(2)

    dataservice_resp = client.get(url_for('api.dataservice_get_specific', dataservice_id=dataservice['id']))
    assert dataservice_resp.status_code == 200
    assert dataservice_resp.json['title'] == dataservice['title']

    dataservice_search_resp = client.get(url_for('api.dataservice_search'))
    assert len(dataservice_search_resp.json['data']) == 1
    assert dataservice_search_resp.json['data'][0]['title'] == dataservice['title']
    assert dataservice_search_resp.json['next_page'] is None
    assert dataservice_search_resp.json['page'] == 1
    assert dataservice_search_resp.json['previous_page'] is None
    assert dataservice_search_resp.json['page_size'] == 20
    assert dataservice_search_resp.json['total_pages'] == 1
    assert dataservice_search_resp.json['total'] == 1

    deletion_resp = client.delete(url_for('api.dataservice_unindex', dataservice_id=dataservice['id']))
    assert deletion_resp.status_code == 200

    time.sleep(2)

    dataservice_get_after_delete_resp = client.get(url_for('api.dataservice_get_specific', dataservice_id=dataservice['id']))
    assert dataservice_get_after_delete_resp.status_code == 404

    dataservice_search_after_delete_resp = client.get(url_for('api.dataservice_search'))
    assert len(dataservice_search_after_delete_resp.json['data']) == 0
    assert dataservice_search_after_delete_resp.json['next_page'] is None
    assert dataservice_search_after_delete_resp.json['page'] == 1
    assert dataservice_search_after_delete_resp.json['previous_page'] is None
    assert dataservice_search_after_delete_resp.json['page_size'] == 20
    assert dataservice_search_after_delete_resp.json['total_pages'] == 1
    assert dataservice_search_after_delete_resp.json['total'] == 0


def test_api_dataservice_index_on_another_index(app, client, search_client, faker):
    now = datetime.datetime.utcnow().strftime('%Y-%m-%d-%H-%M')
    index_name = f"test-dataservice-{now}"
    if not search_client.es.indices.exists(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}"):
        search_client.es.indices.create(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}")

    dataservice = {
        "id": faker.md5(),
        "title": faker.sentence(),
        "description": faker.text(),
        "created_at": faker.past_datetime().isoformat(),
        "followers": faker.random_int(),
        "views": faker.random_int(),
        'organization': {
            'id': faker.md5(),
            'name': faker.company(),
            'public_service': faker.random_int(min=0, max=1),
            'followers': faker.random_int()
        },
        "owner": None,
        "tags": [],
        "is_restricted": False,
        "extras":
        {
            "availability_url": "",
            "is_franceconnect": False,
            "public_cible":
            []
        }
    }

    query = {
        'document': dataservice,
        'index': index_name
    }

    index_resp = client.post(url_for('api.dataservice_index'), json=query)
    assert index_resp.status_code == 200

    time.sleep(2)

    resp = search_client.es.get(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}", id=dataservice['id'])
    assert resp['_source']['title'] == dataservice['title']


def test_api_search_without_query(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory())
        search_client.index_organization(OrganizationFactory())
        search_client.index_reuse(ReuseFactory())

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    dataset_resp = client.get(url_for('api.dataset_search'))
    assert len(dataset_resp.json['data']) == 4
    assert dataset_resp.json['next_page'] is None
    assert dataset_resp.json['page'] == 1
    assert dataset_resp.json['previous_page'] is None
    assert dataset_resp.json['page_size'] == 20
    assert dataset_resp.json['total_pages'] == 1
    assert dataset_resp.json['total'] == 4

    org_resp = client.get(url_for('api.organization_search'))
    assert len(org_resp.json['data']) == 4
    assert org_resp.json['next_page'] is None
    assert org_resp.json['page'] == 1
    assert org_resp.json['previous_page'] is None
    assert org_resp.json['page_size'] == 20
    assert org_resp.json['total_pages'] == 1
    assert org_resp.json['total'] == 4

    reuse_resp = client.get(url_for('api.reuse_search'))
    assert len(reuse_resp.json['data']) == 4
    assert reuse_resp.json['next_page'] is None
    assert reuse_resp.json['page'] == 1
    assert reuse_resp.json['previous_page'] is None
    assert reuse_resp.json['page_size'] == 20
    assert reuse_resp.json['total_pages'] == 1
    assert reuse_resp.json['total'] == 4


def test_api_search_pagination_without_query(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory())
        search_client.index_organization(OrganizationFactory())
        search_client.index_reuse(ReuseFactory())

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    # Test next_url

    dataset_resp = client.get(url_for('api.dataset_search', page_size=2, page=1))
    assert len(dataset_resp.json['data']) == 2
    assert '/api/1/datasets/?page=2&page_size=2' in dataset_resp.json['next_page']
    assert dataset_resp.json['page'] == 1
    assert dataset_resp.json['previous_page'] is None
    assert dataset_resp.json['page_size'] == 2
    assert dataset_resp.json['total_pages'] == 2
    assert dataset_resp.json['total'] == 4

    org_resp = client.get(url_for('api.organization_search', page_size=2, page=1))
    assert len(org_resp.json['data']) == 2
    assert '/api/1/organizations/?page=2&page_size=2' in org_resp.json['next_page']
    assert org_resp.json['page'] == 1
    assert org_resp.json['previous_page'] is None
    assert org_resp.json['page_size'] == 2
    assert org_resp.json['total_pages'] == 2
    assert org_resp.json['total'] == 4

    reuse_resp = client.get(url_for('api.reuse_search', page_size=2, page=1))
    assert len(reuse_resp.json['data']) == 2
    assert '/api/1/reuses/?page=2&page_size=2' in reuse_resp.json['next_page']
    assert reuse_resp.json['page'] == 1
    assert reuse_resp.json['previous_page'] is None
    assert reuse_resp.json['page_size'] == 2
    assert reuse_resp.json['total_pages'] == 2
    assert reuse_resp.json['total'] == 4

    # Test previous_url

    dataset_resp = client.get(url_for('api.dataset_search', page_size=2, page=2))
    assert len(dataset_resp.json['data']) == 2
    assert dataset_resp.json['next_page'] is None
    assert dataset_resp.json['page'] == 2
    assert '/api/1/datasets/?page=1&page_size=2' in dataset_resp.json['previous_page']
    assert dataset_resp.json['page_size'] == 2
    assert dataset_resp.json['total_pages'] == 2
    assert dataset_resp.json['total'] == 4

    org_resp = client.get(url_for('api.organization_search', page_size=2, page=2))
    assert len(org_resp.json['data']) == 2
    assert org_resp.json['next_page'] is None
    assert org_resp.json['page'] == 2
    assert '/api/1/organizations/?page=1&page_size=2' in org_resp.json['previous_page']
    assert org_resp.json['page_size'] == 2
    assert org_resp.json['total_pages'] == 2
    assert org_resp.json['total'] == 4

    reuse_resp = client.get(url_for('api.reuse_search', page_size=2, page=2))
    assert len(reuse_resp.json['data']) == 2
    assert reuse_resp.json['page'] == 2
    assert reuse_resp.json['next_page'] is None
    assert '/api/1/reuses/?page=1&page_size=2' in reuse_resp.json['previous_page']
    assert reuse_resp.json['page_size'] == 2
    assert reuse_resp.json['total_pages'] == 2
    assert reuse_resp.json['total'] == 4


def test_api_create_index(app, client, search_client, faker):
    now = datetime.datetime.utcnow().strftime('%Y-%m-%d-%H-%M')
    index_name = f"test-dataset-{now}"

    payload = {
        'index': index_name
    }
    set_alias_resp = client.post(url_for('api.create_index'), json=payload)
    assert set_alias_resp.status_code == 200

    assert search_client.es.indices.exists(index=f"{app.config['UDATA_INSTANCE_NAME']}-{index_name}")


def test_api_set_index_alias(app, client, search_client, faker):
    now = datetime.datetime.utcnow().strftime('%Y-%m-%d-%H-%M')
    index_name = f"{app.config['UDATA_INSTANCE_NAME']}-dataset-{now}"

    if not search_client.es.indices.exists(index=index_name):
        search_client.es.indices.create(index=index_name)

    payload = {
        'index_suffix_name': now,
        'indices': ['dataset']
    }
    set_alias_resp = client.post(url_for('api.set_index_alias'), json=payload)
    assert set_alias_resp.status_code == 200

    index_alias = f"{app.config['UDATA_INSTANCE_NAME']}-dataset"

    assert search_client.es.indices.exists_alias(name=index_alias)
    alias_keys = list(search_client.es.indices.get_alias(name=index_alias))

    assert alias_keys[0] == index_name


    now = datetime.datetime.utcnow().strftime('%Y-%m-%d-%H-%M')

    for index in ['dataset', 'reuse', 'organization']:
        index_name = f"{app.config['UDATA_INSTANCE_NAME']}-{index}-{now}"
        if not search_client.es.indices.exists(index=index_name):
            search_client.es.indices.create(index=index_name)

    payload = {
        'index_suffix_name': now,
        'indices': []
    }
    set_alias_resp = client.post(url_for('api.set_index_alias'), json=payload)
    assert set_alias_resp.status_code == 200

    for index in ['dataset', 'reuse', 'organization']:

        index_alias = f"{app.config['UDATA_INSTANCE_NAME']}-{index}"
        index_name = f"{app.config['UDATA_INSTANCE_NAME']}-{index}-{now}"

        search_client.es.indices.exists_alias(name=index_alias)
        alias_keys = list(search_client.es.indices.get_alias(name=index_alias))

        assert alias_keys[0] == index_name


def test_api_search_with_query(app, client, search_client, faker):
    for i in range(4):
        title = 'sample-test-{0}'.format(i) if i % 2 else i
        organization = 'sample-test-{0}'.format(faker.company()) if i % 2 else 'not-{0}'.format(faker.company())
        search_client.index_dataset(DatasetFactory(title=title))
        search_client.index_organization(OrganizationFactory(name=organization))
        search_client.index_reuse(ReuseFactory(title='sample-test-{0}'.format(i) if i % 2 else 'not-{0}'.format(i)))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    resp = client.get(url_for('api.dataset_search'))
    assert resp.json['total'] == 4
    resp = client.get(url_for('api.organization_search'))
    assert resp.json['total'] == 4
    resp = client.get(url_for('api.reuse_search'))
    assert resp.json['total'] == 4

    resp = client.get(url_for('api.dataset_search', q='sample-test'))
    assert resp.json['total'] == 2
    resp = client.get(url_for('api.organization_search', q='sample-test'))
    assert resp.json['total'] == 2
    resp = client.get(url_for('api.reuse_search',  q='sample-test'))
    assert resp.json['total'] == 2


def test_api_dataset_search_with_temporal_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            temporal_coverage_start=datetime.date(2021, 12, 2) if i % 2 else datetime.date(2020, 2, 24),
            temporal_coverage_end=datetime.date(2022, 1, 1) if i % 2 else datetime.date(2022, 2, 13)
        ))
    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    resp = client.get(url_for('api.dataset_search', temporal_coverage='2020-02-25-2020-03-10'))
    assert resp.json['total'] == 2

    resp = client.get(url_for('api.dataset_search', temporal_coverage='2020-02-25-20111111111111-10'))
    assert resp.status_code == 400


def test_api_search_with_tag_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(tags=['test-tag', 'test-tag-2'] if i % 2 else ['not-test-tag']))
        search_client.index_reuse(ReuseFactory(tags=['test-tag'] if i % 2 else ['not-test-tag']))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    # Filter is singular, we test the feature that maps it to the plural field of the entity.
    resp = client.get(url_for('api.dataset_search'))
    assert resp.json['total'] == 4

    resp = client.get(url_for('api.dataset_search', tag=['test-tag', 'test-tag-2']))
    assert resp.json['total'] == 2

    resp = client.get(url_for('api.dataset_search', tag=['not-test-tag']))
    assert resp.json['total'] == 2

    resp = client.get(url_for('api.reuse_search'))
    assert resp.json['total'] == 4

    resp = client.get(url_for('api.reuse_search', tag='test-tag'))
    assert resp.json['total'] == 2


def test_api_search_with_topic_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(topics=['test_topic'] if i % 2 else ['not_test_topic']))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    # Filter is singular, we test the feature that maps it to the plural field of the entity.
    resp = client.get(url_for('api.dataset_search'))
    assert resp.json['total'] == 4

    resp = client.get(url_for('api.dataset_search', topic='test_topic'))
    assert resp.json['total'] == 2


def test_api_dataset_search_with_geozone_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            geozones='country:fr' if i % 2 else 'country:ro'
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    resp = client.get(url_for('api.dataset_search'))
    assert resp.json['total'] == 4

    # Filter is singular, we test the feature that maps it to the plural field of the entity.
    resp = client.get(url_for('api.dataset_search', geozone='country:fr'))
    assert resp.json['total'] == 2


def test_api_search_with_followers_sorting(app, client, search_client, faker):
    search_client.index_dataset(DatasetFactory(
        title='data-test-1',
        followers=0
    ))
    search_client.index_dataset(DatasetFactory(
        title='data-test-2',
        followers=3
    ))
    search_client.index_organization(OrganizationFactory(
        name='org-test-1',
        followers=0
    ))
    search_client.index_organization(OrganizationFactory(
        name='org-test-2',
        followers=3
    ))
    search_client.index_reuse(ReuseFactory(
        title='reuse-test-1',
        followers=0
    ))
    search_client.index_reuse(ReuseFactory(
        title='reuse-test-2',
        followers=3
    ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    dataset_resp = client.get(url_for('api.dataset_search', sort='followers'))
    assert len(dataset_resp.json['data']) == 2
    assert dataset_resp.json['data'][0]['title'] == 'data-test-1'
    assert dataset_resp.json['next_page'] is None
    assert dataset_resp.json['page'] == 1
    assert dataset_resp.json['previous_page'] is None
    assert dataset_resp.json['page_size'] == 20
    assert dataset_resp.json['total_pages'] == 1
    assert dataset_resp.json['total'] == 2

    org_resp = client.get(url_for('api.organization_search', sort='followers'))
    assert len(org_resp.json['data']) == 2
    assert org_resp.json['data'][0]['name'] == 'org-test-1'
    assert org_resp.json['next_page'] is None
    assert org_resp.json['page'] == 1
    assert org_resp.json['previous_page'] is None
    assert org_resp.json['page_size'] == 20
    assert org_resp.json['total_pages'] == 1
    assert org_resp.json['total'] == 2

    reuse_resp = client.get(url_for('api.reuse_search', sort='followers'))
    assert len(reuse_resp.json['data']) == 2
    assert reuse_resp.json['data'][0]['title'] == 'reuse-test-1'
    assert reuse_resp.json['next_page'] is None
    assert reuse_resp.json['page'] == 1
    assert reuse_resp.json['previous_page'] is None
    assert reuse_resp.json['page_size'] == 20
    assert reuse_resp.json['total_pages'] == 1
    assert reuse_resp.json['total'] == 2


def test_api_search_with_followers_decreasing_sorting(app, client, search_client, faker):
    search_client.index_dataset(DatasetFactory(
        title='data-test-1',
        followers=0
    ))
    search_client.index_dataset(DatasetFactory(
        title='data-test-2',
        followers=3
    ))
    search_client.index_organization(OrganizationFactory(
        name='org-test-1',
        followers=0
    ))
    search_client.index_organization(OrganizationFactory(
        name='org-test-2',
        followers=3
    ))
    search_client.index_reuse(ReuseFactory(
        title='reuse-test-1',
        followers=0
    ))
    search_client.index_reuse(ReuseFactory(
        title='reuse-test-2',
        followers=3
    ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    dataset_resp = client.get(url_for('api.dataset_search', sort='-followers'))
    assert len(dataset_resp.json['data']) == 2
    assert dataset_resp.json['data'][0]['title'] == 'data-test-2'
    assert dataset_resp.json['next_page'] is None
    assert dataset_resp.json['page'] == 1
    assert dataset_resp.json['previous_page'] is None
    assert dataset_resp.json['page_size'] == 20
    assert dataset_resp.json['total_pages'] == 1
    assert dataset_resp.json['total'] == 2

    org_resp = client.get(url_for('api.organization_search', sort='-followers'))
    assert len(org_resp.json['data']) == 2
    assert org_resp.json['data'][0]['name'] == 'org-test-2'
    assert org_resp.json['next_page'] is None
    assert org_resp.json['page'] == 1
    assert org_resp.json['previous_page'] is None
    assert org_resp.json['page_size'] == 20
    assert org_resp.json['total_pages'] == 1
    assert org_resp.json['total'] == 2

    reuse_resp = client.get(url_for('api.reuse_search', sort='-followers'))
    assert len(reuse_resp.json['data']) == 2
    assert reuse_resp.json['data'][0]['title'] == 'reuse-test-2'
    assert reuse_resp.json['next_page'] is None
    assert reuse_resp.json['page'] == 1
    assert reuse_resp.json['previous_page'] is None
    assert reuse_resp.json['page_size'] == 20
    assert reuse_resp.json['total_pages'] == 1
    assert reuse_resp.json['total'] == 2



import logging
import os

from udata_search_service.domain.entities import Dataset, Organization, Reuse, Dataservice
from udata_search_service.infrastructure.utils import get_concat_title_org, log2p, mdstrip


CONSUMER_LOGGING_LEVEL = int(os.environ.get("CONSUMER_LOGGING_LEVEL", logging.INFO))


class DatasetConsumer(Dataset):
    @classmethod
    def load_from_dict(cls, data):
        # Strip markdown
        data["description"] = mdstrip(data["description"])

        # Retrieve schema renamed as schema_
        data["schema"] = data.get("schema_") or data.get("schema")

        organization = data["organization"]
        data["organization"] = organization.get('id') if organization else None
        data["orga_followers"] = organization.get('followers') if organization else None
        data["orga_sp"] = organization.get('public_service') if organization else None
        data["organization_name"] = organization.get('name') if organization else None
        data["organization_badges"] = organization.get('badges') if organization else None

        resources = data["resources"]
        data["resources_ids"] = [res.get("id") for res in resources]
        data["resources_titles"] = [res.get("title") for res in resources]
        del data["resources"]

        data["concat_title_org"] = get_concat_title_org(data["title"], data['acronym'], data['organization_name'])
        data["geozones"] = [zone.get("id") for zone in data.get("geozones", [])]

        # Normalize values
        data["views"] = log2p(data.get("views", 0))
        data["followers"] = log2p(data.get("followers", 0))
        data["reuses"] = log2p(data.get("reuses", 0))
        data["orga_followers"] = log2p(data.get("orga_followers", 0))
        data["orga_sp"] = 4 if data.get("orga_sp", 0) else 1
        data["featured"] = 4 if data.get("featured", 0) else 1

        return super().load_from_dict(data)


class ReuseConsumer(Reuse):
    @classmethod
    def load_from_dict(cls, data):
        # Strip markdown
        data["description"] = mdstrip(data["description"])

        organization = data["organization"]
        data["organization"] = organization.get('id') if organization else None
        data["orga_followers"] = organization.get('followers') if organization else None
        data["organization_name"] = organization.get('name') if organization else None
        data["organization_badges"] = organization.get('badges') if organization else None

        # Normalize values
        data["views"] = log2p(data.get("views", 0))
        data["followers"] = log2p(data.get("followers", 0))
        data["orga_followers"] = log2p(data.get("orga_followers", 0))
        return super().load_from_dict(data)


class OrganizationConsumer(Organization):
    @classmethod
    def load_from_dict(cls, data):
        # Strip markdown
        data["description"] = mdstrip(data["description"])

        data["followers"] = log2p(data.get("followers", 0))
        data["views"] = log2p(data.get("views", 0))
        return super().load_from_dict(data)


class DataserviceConsumer(Dataservice):
    @classmethod
    def load_from_dict(cls, data):
        # Strip markdown
        data["description"] = mdstrip(data["description"])
        data["description_length"] = len(data["description"])

        organization = data["organization"]
        data["organization"] = organization.get('id') if organization else None
        data["orga_followers"] = organization.get('followers') if organization else None
        data["organization_name"] = organization.get('name') if organization else None

        # Normalize values
        data["views"] = log2p(data.get("views", 0))
        data["followers"] = log2p(data.get("followers", 0))
        data["orga_followers"] = log2p(data.get("orga_followers", 0))
        data["description_length"] = log2p(data["description_length"])

        return super().load_from_dict(data)



import click
from datetime import datetime
from fnmatch import fnmatch
import logging
import os
from typing import Tuple, Optional, List

from elasticsearch import Elasticsearch
from elasticsearch.exceptions import NotFoundError
from elasticsearch_dsl import Date, Document, Float, Integer, Keyword, Text, tokenizer, token_filter, analyzer, query
from elasticsearch_dsl.connections import connections
from udata_search_service.domain.entities import Dataset, Organization, Reuse, Dataservice
from udata_search_service.config import Config
from udata_search_service.infrastructure.utils import IS_TTY

CONSUMER_LOGGING_LEVEL = int(os.environ.get("CONSUMER_LOGGING_LEVEL", logging.INFO))

logging.basicConfig(level=CONSUMER_LOGGING_LEVEL)


# Définition d'un analyzer français (repris ici : https://jolicode.com/blog/construire-un-bon-analyzer-francais-pour-elasticsearch)
# Ajout dans la filtre french_synonym, des synonymes que l'on souhaite implémenter (ex : AMD / Administrateur des Données)
# Création du mapping en indiquant les champs sur lesquels cet analyzer va s'appliquer (title, description, concat, organization)
# et en spécifiant les types de champs que l'on va utiliser pour calculer notre score de pertinence
french_elision = token_filter('french_elision', type='elision', articles_case=True, articles=["l", "m", "t", "qu", "n", "s", "j", "d", "c", "jusqu", "quoiqu", "lorsqu", "puisqu"])
french_stop = token_filter('french_stop', type='stop', stopwords='_french_')
french_stemmer = token_filter('french_stemmer', type='stemmer', language='light_french')
french_synonym = token_filter('french_synonym', type='synonym', ignore_case=True, expand=True, synonyms=Config.SEARCH_SYNONYMS)


dgv_analyzer = analyzer('french_dgv',
                        tokenizer=tokenizer('icu_tokenizer'),
                        filter=['icu_folding', french_elision, french_synonym, french_stemmer, french_stop]
                        )


class IndexDocument(Document):

    @classmethod
    def init_index(cls, es_client: Elasticsearch, suffix: str) -> None:
        alias = cls._index._name
        pattern = alias + '-*'

        logging.info(f'Saving template {alias} on the following pattern: {pattern}')
        index_template = cls._index.as_template(alias, pattern)
        index_template.save()

        if not cls._index.exists():
            logging.info(f'Creating index {alias + suffix}')
            es_client.indices.create(index=alias + suffix)
            es_client.indices.put_alias(index=alias + suffix, name=alias)
        else:
            logging.info(f'Index on alias {alias} already exists')

    @classmethod
    def delete_indices(cls, es_client: Elasticsearch) -> None:
        alias = cls._index._name
        pattern = alias + '*'
        logging.info(f'Deleting indices with pattern {pattern}')
        es_client.indices.delete(index=pattern)

    @classmethod
    def _matches(cls, hit):
        # override _matches to match indices in a pattern instead of just ALIAS
        # hit is the raw dict as returned by elasticsearch
        alias = cls._index._name
        pattern = alias + '-*'
        return fnmatch(hit["_index"], pattern)


class SearchableDataservice(IndexDocument):
    title = Text(analyzer=dgv_analyzer)
    created_at = Date()
    tags = Keyword(multi=True)
    organization = Keyword()
    description = Text(analyzer=dgv_analyzer)
    organization_name = Text(analyzer=dgv_analyzer)
    owner = Keyword()
    views = Float()
    followers = Float()
    description_length = Float()

    class Index:
        name = f'{Config.UDATA_INSTANCE_NAME}-dataservice'


class SearchableOrganization(IndexDocument):
    name = Text(analyzer=dgv_analyzer)
    acronym = Text()
    description = Text(analyzer=dgv_analyzer)
    url = Text()
    orga_sp = Integer()
    created_at = Date()
    followers = Float()
    views = Float()
    reuses = Float()
    datasets = Integer()
    badges = Keyword(multi=True)

    class Index:
        name = f'{Config.UDATA_INSTANCE_NAME}-organization'


class SearchableReuse(IndexDocument):
    title = Text(analyzer=dgv_analyzer)
    url = Text()
    created_at = Date()
    archived = Date()
    orga_followers = Float()
    views = Float()
    followers = Float()
    datasets = Integer()
    featured = Integer()
    type = Keyword()
    topic = Keyword()
    tags = Keyword(multi=True)
    badges = Keyword(multi=True)
    organization = Keyword()
    description = Text(analyzer=dgv_analyzer)
    organization_name = Text(analyzer=dgv_analyzer)
    organization_badges = Keyword(multi=True)
    owner = Keyword()

    class Index:
        name = f'{Config.UDATA_INSTANCE_NAME}-reuse'


class SearchableDataset(IndexDocument):
    title = Text(analyzer=dgv_analyzer)
    acronym = Text()
    url = Text()
    created_at = Date()
    last_update = Date()
    tags = Keyword(multi=True)
    license = Keyword()
    badges = Keyword(multi=True)
    frequency = Text()
    format = Keyword(multi=True)
    orga_sp = Integer()
    orga_followers = Float()
    views = Float()
    followers = Float()
    reuses = Float()
    featured = Integer()
    resources_count = Integer()
    resources_ids = Keyword(multi=True)
    resources_titles = Text(analyzer=dgv_analyzer)
    concat_title_org = Text(analyzer=dgv_analyzer)
    temporal_coverage_start = Date()
    temporal_coverage_end = Date()
    granularity = Keyword()
    geozones = Keyword(multi=True)
    description = Text(analyzer=dgv_analyzer)
    organization = Keyword()
    organization_name = Text(analyzer=dgv_analyzer)
    organization_badges = Keyword(multi=True)
    owner = Keyword()
    schema = Keyword(multi=True)
    topics = Keyword(multi=True)

    class Index:
        name = f'{Config.UDATA_INSTANCE_NAME}-dataset'


class ElasticClient:

    def __init__(self, url: str):
        self.es = connections.create_connection(hosts=[url])

    def init_indices(self) -> None:
        '''
        Create templates based on Document mappings and map patterns.
        Create time-based index matchin the template patterns.
        '''
        suffix_name = '-' + datetime.utcnow().strftime('%Y-%m-%d-%H-%M')

        SearchableDataset.init_index(self.es, suffix_name)
        SearchableReuse.init_index(self.es, suffix_name)
        SearchableOrganization.init_index(self.es, suffix_name)
        SearchableDataservice.init_index(self.es, suffix_name)

    def clean_indices(self) -> None:
        '''
        Removing previous indices and intializing new ones.
        '''

        if IS_TTY:
            msg = 'Indices will be deleted, are you sure?'
            click.confirm(msg, abort=True)
        SearchableDataset.delete_indices(self.es)
        SearchableReuse.delete_indices(self.es)
        SearchableOrganization.delete_indices(self.es)
        SearchableDataservice.delete_indices(self.es)

        self.init_indices()

    def index_organization(self, to_index: Organization, index: str = None) -> None:
        SearchableOrganization(meta={'id': to_index.id}, **to_index.to_dict()).save(skip_empty=False, index=index)

    def index_dataset(self, to_index: Dataset, index: str = None) -> None:
        SearchableDataset(meta={'id': to_index.id}, **to_index.to_dict()).save(skip_empty=False, index=index)

    def index_reuse(self, to_index: Reuse, index: str = None) -> None:
        SearchableReuse(meta={'id': to_index.id}, **to_index.to_dict()).save(skip_empty=False, index=index)

    def index_dataservice(self, to_index: Dataservice, index: str = None) -> None:
        SearchableDataservice(meta={'id': to_index.id}, **to_index.to_dict()).save(skip_empty=False, index=index)

    def query_organizations(self, query_text: str, offset: int, page_size: int, filters: dict, sort: Optional[str] = None) -> Tuple[int, List[dict]]:
        search = SearchableOrganization.search()

        for key, value in filters.items():
            search = search.filter('term', **{key: value})

        organizations_score_functions = [
            query.SF("field_value_factor", field="orga_sp", factor=8, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="followers", factor=4, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="views", factor=1, modifier='sqrt', missing=1),
        ]

        if query_text:
            search = search.query('bool', should=[
                    query.Q(
                        'function_score',
                        query=query.Bool(should=[query.MultiMatch(query=query_text, type='phrase', fields=['id^15', 'name^15', 'acronym^15', 'description^8'])]),
                        functions=organizations_score_functions
                    ),
                    query.Q(
                        'function_score',
                        query=query.Bool(should=[query.MultiMatch(
                            query=query_text,
                            type='cross_fields',
                            fields=['id^15', 'name^7', 'acronym^7', 'description^4'],
                            operator="and")]),
                        functions=organizations_score_functions
                    ),
                    query.Match(title={"query": query_text, 'fuzziness': 'AUTO:4,6'}),
            ])
        else:
            search = search.query(query.Q('function_score', query=query.MatchAll(), functions=organizations_score_functions))

        if sort:
            search = search.sort(sort, {'_score': {'order': 'desc'}})

        search = search[offset:(offset + page_size)]

        response = search.execute()
        results_number = response.hits.total.value
        if response.hits and not isinstance(response.hits[0], SearchableOrganization):
            raise ValueError(
                'Results are not of SearchableOrganization type. It probably means that index analyzers '
                'were not correctly set using template patterns on index initialization.'
            )
        res = [hit.to_dict(skip_empty=False) for hit in response.hits]
        return results_number, res

    def query_datasets(self, query_text: str, offset: int, page_size: int, filters: dict, sort: Optional[str] = None) -> Tuple[int, List[dict]]:
        search = SearchableDataset.search()

        for key, value in filters.items():
            if key == 'temporal_coverage_start':
                search = search.filter('range', **{'temporal_coverage_start': {'lte': value}})
            elif key == 'temporal_coverage_end':
                search = search.filter('range', **{'temporal_coverage_end': {'gte': value}})
            elif key == 'tags':
                # build an AND filter from tags list
                tag_filters = [query.Q('term', tags=tag) for tag in value]
                search = search.filter(
                    query.Bool(must=tag_filters)
                )
            else:
                search = search.filter('term', **{key: value})

        datasets_score_functions = [
            query.SF("field_value_factor", field="orga_sp", factor=8, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="views", factor=4, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="followers", factor=4, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="orga_followers", factor=1, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="featured", factor=1, modifier='sqrt', missing=1),
        ]

        if query_text:
            search = search.query(
                'bool',
                should=[
                    query.Q(
                        'function_score',
                        query=query.Bool(should=[query.MultiMatch(
                            query=query_text,
                            type='phrase',
                            fields=['id^15', 'title^15', 'acronym^15', 'description^8', 'organization_name^8', 'resources_ids^8', 'resources_titles^5']
                        )]),
                        functions=datasets_score_functions
                    ),
                    query.Q(
                        'function_score',
                        query=query.Bool(must=[query.Match(concat_title_org={"query": query_text, "operator": "and", "boost": 8})]),
                        functions=datasets_score_functions,
                    ),
                    query.Q(
                        'function_score',
                        query=query.Bool(should=[query.MultiMatch(
                            query=query_text,
                            type='cross_fields',
                            fields=['id^7', 'title^7', 'acronym^7', 'description^4', 'organization_name^4', 'resources_ids^4', 'resources_titles^2'],
                            operator="and")]),
                        functions=datasets_score_functions
                    ),
                    query.MultiMatch(query=query_text, type='most_fields', operator="and", fields=['title', 'organization_name'], fuzziness='AUTO:4,6')
                ])
        else:
            search = search.query(query.Q('function_score', query=query.MatchAll(), functions=datasets_score_functions))

        if sort:
            search = search.sort(sort, {'_score': {'order': 'desc'}})

        search = search[offset:(offset + page_size)]

        response = search.execute()
        results_number = response.hits.total.value
        if response.hits and not isinstance(response.hits[0], SearchableDataset):
            raise ValueError(
                'Results are not of SearchableDataset type. It probably means that index analyzers were not correctly set '
                'using template patterns on index initialization.'
            )
        res = [hit.to_dict(skip_empty=False) for hit in response.hits]
        return results_number, res

    def query_reuses(self, query_text: str, offset: int, page_size: int, filters: dict, sort: Optional[str] = None) -> Tuple[int, List[dict]]:
        search = SearchableReuse.search()

        for key, value in filters.items():
            search = search.filter('term', **{key: value})

        reuses_score_functions = [
            query.SF("field_value_factor", field="views", factor=4, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="followers", factor=4, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="orga_followers", factor=1, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="featured", factor=1, modifier='sqrt', missing=1),
            query.SF("script_score", script={"source": "doc['archived'].size() == 0 ? 1 : 0.2"}),  # score is multiplied by 0.2 for a strong malus
        ]

        if query_text:
            search = search.query('bool', should=[
                    query.Q(
                        'function_score',
                        query=query.Bool(should=[query.MultiMatch(query=query_text, type='phrase', fields=['id^15', 'title^15', 'description^8', 'organization_name^8'])]),
                        functions=reuses_score_functions
                    ),
                    query.Q(
                        'function_score',
                        query=query.Bool(should=[query.MultiMatch(
                            query=query_text,
                            type='cross_fields',
                            fields=['id^7', 'title^7', 'description^4', 'organization_name^4'],
                            operator="and")]),
                        functions=reuses_score_functions
                    ),
                    query.MultiMatch(query=query_text, type='most_fields', operator="and", fields=['title', 'organization_name'], fuzziness='AUTO:4,6')
                ])
        else:
            search = search.query(query.Q('function_score', query=query.MatchAll(), functions=reuses_score_functions))

        if sort:
            search = search.sort(sort, {'_score': {'order': 'desc'}})

        search = search[offset:(offset + page_size)]

        response = search.execute()
        results_number = response.hits.total.value
        if response.hits and not isinstance(response.hits[0], SearchableReuse):
            raise ValueError(
                'Results are not of SearchableReuse type. It probably means that index analyzers were not correctly set '
                'using template patterns on index initialization.'
            )
        res = [hit.to_dict(skip_empty=False) for hit in response.hits]
        return results_number, res

    def query_dataservices(self, query_text: str, offset: int, page_size: int, filters: dict, sort: Optional[str] = None) -> Tuple[int, List[dict]]:
        search = SearchableDataservice.search()

        for key, value in filters.items():
            search = search.filter('term', **{key: value})

        dataservices_score_functions = [
            query.SF("field_value_factor", field="description_length", factor=1, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="views", factor=4, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="followers", factor=4, modifier='sqrt', missing=1),
            query.SF("field_value_factor", field="orga_followers", factor=1, modifier='sqrt', missing=1),
        ]

        if query_text:
            search = search.query('bool', should=[
                    query.Q(
                        'function_score',
                        query=query.Bool(should=[query.MultiMatch(query=query_text, type='phrase', fields=['id^15', 'title^15', 'description^8', 'organization_name^8'])]),
                        functions=dataservices_score_functions
                    ),
                    query.Q(
                        'function_score',
                        query=query.Bool(should=[query.MultiMatch(
                            query=query_text,
                            type='cross_fields',
                            fields=['id^7', 'title^7', 'description^4', 'organization_name^4'],
                            operator="and")]),
                        functions=dataservices_score_functions
                    ),
                    query.MultiMatch(query=query_text, type='most_fields', operator="and", fields=['title', 'organization_name'], fuzziness='AUTO:4,6')
                ])
        else:
            search = search.query(query.Q('function_score', query=query.MatchAll(), functions=dataservices_score_functions))

        if sort:
            search = search.sort(sort, {'_score': {'order': 'desc'}})

        search = search[offset:(offset + page_size)]

        response = search.execute()
        results_number = response.hits.total.value
        if response.hits and not isinstance(response.hits[0], SearchableDataservice):
            raise ValueError(
                'Results are not of SearchableDataservice type. It probably means that index analyzers were not correctly set '
                'using template patterns on index initialization.'
            )
        res = [hit.to_dict(skip_empty=False) for hit in response.hits]
        return results_number, res

    def find_one_organization(self, organization_id: str) -> Optional[dict]:
        try:
            return SearchableOrganization.get(id=organization_id).to_dict()
        except NotFoundError:
            return None

    def find_one_dataset(self, dataset_id: str) -> Optional[dict]:
        try:
            return SearchableDataset.get(id=dataset_id).to_dict()
        except NotFoundError:
            return None

    def find_one_reuse(self, reuse_id: str) -> Optional[dict]:
        try:
            return SearchableReuse.get(id=reuse_id).to_dict()
        except NotFoundError:
            return None

    def find_one_dataservice(self, dataservice_id: str) -> Optional[dict]:
        try:
            return SearchableDataservice.get(id=dataservice_id).to_dict()
        except NotFoundError:
            return None

    def delete_one_organization(self, organization_id: str) -> Optional[str]:
        try:
            SearchableOrganization.get(id=organization_id).delete()
            return organization_id
        except NotFoundError:
            return None

    def delete_one_dataset(self, dataset_id: str) -> Optional[str]:
        try:
            SearchableDataset.get(id=dataset_id).delete()
            return dataset_id
        except NotFoundError:
            return None

    def delete_one_reuse(self, reuse_id: str) -> Optional[str]:
        try:
            SearchableReuse.get(id=reuse_id).delete()
            return reuse_id
        except NotFoundError:
            return None

    def delete_one_dataservice(self, dataservice_id: str) -> Optional[str]:
        try:
            SearchableDataservice.get(id=dataservice_id).delete()
            return dataservice_id
        except NotFoundError:
            return None



import click
from flask.cli import with_appcontext
from dependency_injector.wiring import inject, Provide
from udata_search_service.container import Container
from udata_search_service.domain.interfaces import SearchClient


@inject
def init_es_func(search_client: SearchClient = Provide[Container.search_client]) -> None:
    click.echo("Setting up indices and aliases if they don't exist.")
    search_client.init_indices()
    click.echo("Done.")


@inject
def clean_es_func(search_client: SearchClient = Provide[Container.search_client]) -> None:
    click.echo("Removing previous indices and intializing new ones.")
    search_client.clean_indices()
    click.echo("Done.")


@click.command('init-es')
@with_appcontext
def init_es() -> None:
    init_es_func()


@click.command("clean-es")
@with_appcontext
def clean_es_command() -> None:
    clean_es_func()


def init_app(cli):
    cli.add_command(init_es)
    cli.add_command(clean_es_command)



import copy
import datetime

from udata_search_service.infrastructure.consumers import ReuseConsumer, OrganizationConsumer, DataserviceConsumer, DatasetConsumer
from udata_search_service.infrastructure.utils import get_concat_title_org, log2p, mdstrip


def test_parse_dataset_obj():
    obj = {
        'id': '5c4ae55a634f4117716d5656',
        'title': 'Demandes de valeurs foncières',
        'description': '### Propos liminaires...',
        'acronym': 'DVF',
        'url': '/fr/datasets/demandes-de-valeurs-foncieres/',
        'tags': ['foncier', 'foncier-sol-mutation-fonciere', 'fonciere', 'valeur-fonciere'],
        'license': 'notspecified',
        'badges': [],
        'frequency': 'semiannual',
        'created_at': '2019-01-25T11:30:50',
        'last_update': '2020-02-27T12:32:50',
        'views': 7806,
        'followers': 72,
        'reuses': 45,
        'featured': 0,
        'resources_count': 10,
        'resources': [
            {"id": "5ffa8553-0e8f-4622-add9-5c0b593ca1f8", "title": "Valeurs foncières 2024"},
            {"id": "bc213c7c-c4d4-4385-bf1f-719573d39e90", "title": "Valeurs foncières 2023"},
        ],
        'organization': {
            'id': '534fff8ea3a7292c64a77f02',
            'name': 'Ministère de l\'économie, des finances et de la relance',
            'public_service': 1,
            'followers': 401,
            'badges': ["public-service"]},
        'owner': None,
        'format': ['pdf', 'pdf', 'pdf', 'pdf', 'txt', 'txt', 'txt', 'txt', 'txt', 'txt'],
        'temporal_coverage_start': '2016-07-01T00:00:00',
        'temporal_coverage_end': '2021-06-30T00:00:00',
        'geozones': [{'id': 'fr:arrondissement:353', 'name': 'Rennes', 'keys': ['353']},
                     {'id': 'country-group:world'},
                     {'id': 'country:fr'},
                     {'id': 'country-group:ue'}],
        'granularity': 'fr:commune',
        'schema_': ['etalab/schema-irve']
    }
    document = DatasetConsumer.load_from_dict(copy.deepcopy(obj)).to_dict()

    # Make sure that these fields are loaded as is
    for key in ['id', 'title', 'url', 'frequency', 'resources_count',
                'acronym', 'badges', 'tags', 'license', 'owner']:
        assert document[key] == obj[key]

    # Make sure that markdown fields are stripped
    assert document["description"] == mdstrip(obj["description"])

    # Make sure schema keyword is retrieved correctly
    assert document["schema"] == obj["schema_"]

    # Make sure that these fields are log2p-normalized
    for key in ['views', 'followers', 'reuses']:
        assert document[key] == log2p(obj[key])

    # Make sure that all other particular fields are treated accordingly
    assert document['concat_title_org'] == get_concat_title_org(document['title'], document['acronym'], document['organization_name'])
    assert document['created_at'].date() == datetime.date(2019, 1, 25)
    assert document['last_update'].date() == datetime.date(2020, 2, 27)
    assert document['temporal_coverage_start'].date() == datetime.date(2016, 7, 1)
    assert document['temporal_coverage_end'].date() == datetime.date(2021, 6, 30)
    assert document['granularity'] == 'fr:commune'
    assert document['geozones'] == ['fr:arrondissement:353', 'country-group:world', 'country:fr', 'country-group:ue']
    assert document['organization'] == obj['organization']['id']
    assert document['organization_name'] == obj['organization']['name']
    assert document['organization_badges'] == obj['organization']['badges']
    assert document['orga_followers'] == log2p(401)
    assert document['resources_ids'] == [res["id"] for res in obj["resources"]]
    assert document['resources_titles'] == [res["title"] for res in obj["resources"]]


def test_parse_reuse_obj():
    obj = {
        "id": "5cc2dfbe8b4c414c91ffc46d",
        "title": "Explorateur de données de valeur foncière (DVF)",
        "description": "Cartographie des mutations à titre onéreux (parcelles en bleu).",
        "url": "https://app.dvf.etalab.gouv.fr/",
        "created_at": "2019-04-26T12:38:54",
        "views": 4326,
        "followers": 11,
        "datasets": 2,
        "featured": 1,
        "organization": {
            "id": "534fff75a3a7292c64a77de4",
            "name": "Etalab",
            "public_service": 1,
            "followers": 357,
            'badges': ["public-service"]},
        "owner": None,
        "type": "application",
        "topic": "housing_and_development",
        "tags": ["application-cartographique", "cadastre", "dgfip", "dvf", "etalab", "foncier", "mutations"],
        "badges": []
    }
    document = ReuseConsumer.load_from_dict(copy.deepcopy(obj)).to_dict()

    # Make sure that these fields are loaded as is
    for key in ['id', 'title', 'url', 'datasets', 'featured',
                'badges', 'tags', 'owner']:
        assert document[key] == obj[key]

    # Make sure that markdown fields are stripped
    assert document["description"] == mdstrip(obj["description"])

    # Make sure that these fields are log2p-normalized
    for key in ['views', 'followers']:
        assert document[key] == log2p(obj[key])

    # Make sure that all other particular fields are treated accordingly
    assert document['created_at'].date() == datetime.date(2019, 4, 26)
    assert document['organization'] == obj['organization']['id']
    assert document['organization_name'] == obj['organization']['name']
    assert document['organization_badges'] == obj['organization']['badges']
    assert document['orga_followers'] == log2p(obj['organization']['followers'])


def test_parse_dataservice_obj():
    obj = {
        "id": "670786ec0e65e6a0c784bf24",
        "title": "API INES",
        "description": "### À quoi sert l'API INES ?\n\nL’API INES (Identifiant National dans l'Enseignement Supérieur) permet la vérification et l’immatriculation des étudiants dans l’enseignement supérieur.\nLa vérification s’effectue à partir des données d’état civil de l’étudiant et permet de récupérer son identifiant national étudiant unique s’il est déjà immatriculé.\nL’immatriculation permet d’obtenir un identifiant national unique pour l’étudiant, qui lui sera associé durant tout son parcours d’étudiant.\n\nINES permet une immatriculation unique de tout étudiant:\n- dans l’ensemble des établissements, indépendamment de ses mobilités\n- dans tous les autres systèmes d’information sur les étudiants\nL'identifiant national de l’étudiant est une donnée pivot pour mettre en relation l’ensemble des SI les concernant, qu’ils émanent des établissements, du Ministère de l'Enseignement Supérieur et de la Recherche et des administrations publics (StatutEtudiant, remontées SISE), ou des opérateurs (PARCOURSUP, CNOUS…).\n\n\n## Principales données disponibles\n\nL’API s’adresse à tous les établissements proposant une formation dans l’enseignement supérieur et aux systèmes d’information dans l'enseignement supérieur, comme les plateformes de candidature aux établissements d’enseignement supérieur ou le CNOUS.\n\nL’obtention d’un identifiant national étudiant (INE) d’INES se fait à partir des données d’état civil de l’étudiant.\n\nUne authentification (identifiants d’accès directement fourni par le MESR aux établissements) est nécessaire. Sinon, il n’y a pas de restriction particulière quant à l’utilisation de cette API.\n\n[Lien des documentations supplémentaires](https://ines.enseignementsup-recherche.gouv.fr)",
        "created_at": "2024-10-10T07:49:00",
        "views": 4326,
        "followers": 0,
        "organization":
        {
            "id": "534fff90a3a7292c64a77f45",
            "name": "Ministère de l'Enseignement supérieur et de la Recherche",
            "public_service": 1,
            "followers": 152
        },
        "owner": None,
        "tags":
        [
            "education",
            "enseignement",
            "identifiant",
            "immatriculation",
            "ines",
            "verification"
        ],
        "is_restricted": False,
        "extras":
        {
            "availability_url": "",
            "is_franceconnect": False,
            "public_cible":
            []
        }
    }
    document = DataserviceConsumer.load_from_dict(copy.deepcopy(obj)).to_dict()

    # Make sure that these fields are loaded as is
    for key in ['id', 'title', 'tags', 'owner', 'is_restricted']:
        assert document[key] == obj[key]

    # Make sure that markdown fields are stripped
    assert document["description"] == mdstrip(obj["description"])

    # Make sure that these fields are log2p-normalized
    for key in ['views', 'followers']:
        assert document[key] == log2p(obj[key])

    # Make sure that all other particular fields are treated accordingly
    assert document['created_at'].date() == datetime.date(2024, 10, 10)
    assert document['organization'] == obj['organization']['id']
    assert document['organization_name'] == obj['organization']['name']
    assert document['orga_followers'] == log2p(obj['organization']['followers'])
    assert document['description_length'] == log2p(len(mdstrip(obj['description'])))


def test_parse_organization_obj():
    obj = {
        "id": "534fff75a3a7292c64a77de4",
        "name": "Etalab",
        "acronym": None,
        "description": "Etalab est un département de la direction interministérielle du numérique (DINUM)",
        "url": "https://www.etalab.gouv.fr",
        "badges": ["public-service", "certified"],
        "created_at": "2014-04-17T18:21:09",
        "orga_sp": 1,
        "followers": 357,
        "datasets": 56,
        "views": 42,
        "reuses": 0
    }
    document = OrganizationConsumer.load_from_dict(copy.deepcopy(obj)).to_dict()

    # Make sure that these fields are loaded as is
    for key in ['id', 'name', 'acronym', 'url', 'badges', 'orga_sp', 'datasets', 'reuses']:
        assert document[key] == obj[key]

    # Make sure that markdown fields are stripped
    assert document["description"] == mdstrip(obj["description"])

    # Make sure that these fields are log2p-normalized
    assert document["followers"] == log2p(obj["followers"])
    assert document["views"] == log2p(obj["views"])

    # Make sure that all other particular fields are treated accordingly
    assert document['created_at'].date() == datetime.date(2014, 4, 17)



from typing import Type, get_origin, Union, get_args


def is_list_type(type_: Type) -> bool:
    origin = get_origin(type_)
    if origin is list:
        return True
    if origin is Union:
        return any(is_list_type(t) for t in get_args(type_))
    return False



import datetime
import time

from udata_search_service.domain.factories import DataserviceFactory, DatasetFactory, OrganizationFactory, ReuseFactory

ext_word_list = ['abc', 'def', 'hij', 'klm', 'nop', 'qrs', 'tuv']


def test_general_search_with_and_without_query(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            title='test-{0}'.format(i) if i % 2 else faker.word(ext_word_list=ext_word_list),
            description='udata' if i == 1 else faker.word()
        ))
        search_client.index_organization(OrganizationFactory(
            name='test-{0}'.format(faker.company()) if i % 2 else faker.company(),
            description='udata' if i == 1 else faker.word()
        ))
        search_client.index_reuse(ReuseFactory(
            title='test-{0}'.format(i) if i % 2 else faker.word(ext_word_list=ext_word_list),
            description='udata' if i == 1 else faker.word()
        ))
        search_client.index_dataservice(DataserviceFactory(
            title='test-{0}'.format(i) if i % 2 else faker.word(ext_word_list=ext_word_list),
            description='udata' if i == 1 else faker.word()
        ))
    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    # Should return only the object with string 'test' in their titles/names
    results_number, res = search_client.query_datasets('test', 0, 20, {})
    assert results_number == 2
    results_number, res = search_client.query_organizations('test', 0, 20, {})
    assert results_number == 2
    results_number, res = search_client.query_reuses('test', 0, 20, {})
    assert results_number == 2
    results_number, res = search_client.query_dataservices('test', 0, 20, {})
    assert results_number == 2

    # Should return only the object with string 'test' and 'udata' in their titles/names or desc
    results_number, res = search_client.query_datasets('test udata', 0, 20, {})
    assert results_number == 1
    results_number, res = search_client.query_organizations('test udata', 0, 20, {})
    assert results_number == 1
    results_number, res = search_client.query_reuses('test udata', 0, 20, {})
    assert results_number == 1
    results_number, res = search_client.query_dataservices('test udata', 0, 20, {})
    assert results_number == 1

    # Should return all objects, as query text is none
    results_number, res = search_client.query_datasets(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_organizations(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_reuses(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_dataservices(None, 0, 20, {})
    assert results_number == 4


def test_search_with_orga_sp_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            orga_sp=4 if i % 2 else 1
        ))
        search_client.index_organization(OrganizationFactory(
            orga_sp=4 if i % 2 else 1
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_datasets(None, 0, 20, {'orga_sp': 4})
    assert results_number == 2
    results_number, res = search_client.query_organizations(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_organizations(None, 0, 20, {'orga_sp': 4})
    assert results_number == 2


def test_search_with_orga_id_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory())
        search_client.index_reuse(ReuseFactory())

    search_client.index_dataset(DatasetFactory(
        organization='64f01c248bf99eab7c197717'
    ))
    search_client.index_reuse(ReuseFactory(
        organization='77f01c346bf99eab7c198891'
    ))
    search_client.index_dataservice(DataserviceFactory(
        organization='77f01c346bf99eab7c198891'
    ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(None, 0, 20, {})
    assert results_number == 5
    results_number, res = search_client.query_datasets(None, 0, 20, {'organization': '64f01c248bf99eab7c197717'})
    assert results_number == 1
    results_number, res = search_client.query_reuses(None, 0, 20, {})
    assert results_number == 5
    results_number, res = search_client.query_reuses(None, 0, 20, {'organization': '77f01c346bf99eab7c198891'})
    assert results_number == 1
    results_number, res = search_client.query_dataservices(None, 0, 20, {'organization': '77f01c346bf99eab7c198891'})
    assert results_number == 1


def test_search_with_orga_badge_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            organization_badges=["public-service"] if i % 2 else []
        ))
        search_client.index_reuse(ReuseFactory(
            organization_badges=["public-service"] if i % 2 else []
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_datasets(None, 0, 20, {'organization_badges': 'public-service'})
    assert results_number == 2
    results_number, res = search_client.query_reuses(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_reuses(None, 0, 20, {'organization_badges': 'public-service'})
    assert results_number == 2


def test_search_with_owner_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory())
        search_client.index_reuse(ReuseFactory())

    search_client.index_dataset(DatasetFactory(
        owner='64f01c248bf99eab7c197717'
    ))

    search_client.index_reuse(ReuseFactory(
        owner='77f01c346bf99eab7c198891'
    ))

    search_client.index_dataservice(DataserviceFactory(
        owner='77f01c346bf99eab7c198891'
    ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(None, 0, 20, {})
    assert results_number == 5
    results_number, res = search_client.query_datasets(None, 0, 20, {'owner': '64f01c248bf99eab7c197717'})
    assert results_number == 1
    results_number, res = search_client.query_reuses(None, 0, 20, {})
    assert results_number == 5
    results_number, res = search_client.query_reuses(None, 0, 20, {'owner': '77f01c346bf99eab7c198891'})
    assert results_number == 1
    results_number, res = search_client.query_dataservices(None, 0, 20, {'owner': '77f01c346bf99eab7c198891'})
    assert results_number == 1


def test_search_datasets_with_temporal_filters(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            temporal_coverage_start=datetime.date(2021, 12, 2) if i % 2 else datetime.date(2020, 2, 24),
            temporal_coverage_end=datetime.date(2022, 1, 1) if i % 2 else datetime.date(2022, 2, 13)
        ))
    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_datasets(None, 0, 20, {
        'temporal_coverage_start': datetime.date(2020, 2, 25),
        'temporal_coverage_end': datetime.date(2020, 3, 10)
    })
    assert results_number == 2
    results_number, res = search_client.query_datasets(None, 0, 20, {
        'temporal_coverage_start': datetime.date(2019, 2, 25),
        'temporal_coverage_end': datetime.date(2020, 3, 10)
    })
    assert results_number == 0


def test_search_with_tag_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            tags=['test-tag', f'test-tag-{i}'] if i % 2 else ['not-test-tag']
        ))

        search_client.index_reuse(ReuseFactory(
            tags=['test-tag'] if i % 2 else ['not-test-tag']
        ))

        search_client.index_dataservice(DataserviceFactory(
            tags=['test-tag'] if i % 2 else ['not-test-tag']
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_datasets(None, 0, 20, {'tags': ['test-tag', 'test-tag-1']})
    assert results_number == 1
    results_number, res = search_client.query_datasets(None, 0, 20, {'tags': ['not-test-tag']})
    assert results_number == 2
    results_number, res = search_client.query_reuses(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_reuses(None, 0, 20, {'tags': 'test-tag'})
    assert results_number == 2
    results_number, res = search_client.query_dataservices(None, 0, 20, {'tags': 'test-tag'})
    assert results_number == 2


def test_search_dataset_with_format_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            format=['pdf'] if i % 2 else ['csv']
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_datasets(None, 0, 20, {'format': 'pdf'})
    assert results_number == 2


def test_search_dataset_with_license_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            license='cc-by' if i % 2 else 'fr-lo'
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_datasets(None, 0, 20, {'license': 'cc-by'})
    assert results_number == 2


def test_search_dataset_with_geozone_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            geozones='country:fr' if i % 2 else 'country:ro'
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_datasets(None, 0, 20, {'geozones': 'country:fr'})
    assert results_number == 2


def test_search_dataset_with_granularity_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            granularity='country' if i % 2 else 'country-subset'
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_datasets(None, 0, 20, {'granularity': 'country'})
    assert results_number == 2


def test_search_dataset_with_schema_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataset(DatasetFactory(
            title='data-test-1',
            schema=['etalab/schema-irve'] if i % 2 else []
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets('test', 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_datasets('test', 0, 20, {'schema': 'etalab/schema-irve'})
    assert results_number == 2


def test_search_reuse_with_type_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_reuse(ReuseFactory(
            type='api' if i % 2 else 'application'
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_reuses(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_reuses(None, 0, 20, {'type': 'api'})
    assert results_number == 2


def test_search_reuse_with_topic_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_reuse(ReuseFactory(
            topic='transport_and_mobility' if i % 2 else 'housing_and_development'
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_reuses(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_reuses(None, 0, 20, {'topic': 'transport_and_mobility'})
    assert results_number == 2


def test_search_dataservice_with_is_restricted_filter(app, client, search_client, faker):
    for i in range(4):
        search_client.index_dataservice(DataserviceFactory(
            is_restricted=bool(i % 2)
        ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_dataservices(None, 0, 20, {})
    assert results_number == 4
    results_number, res = search_client.query_dataservices(None, 0, 20, {'is_restricted': True})
    assert results_number == 2
    results_number, res = search_client.query_dataservices(None, 0, 20, {'is_restricted': False})
    assert results_number == 2


def test_general_search_with_sorting(app, client, search_client, faker):
    search_client.index_dataset(DatasetFactory(
        title='data-test-1',
        followers=0
    ))
    search_client.index_dataset(DatasetFactory(
        title='data-test-2',
        followers=3
    ))
    search_client.index_organization(OrganizationFactory(
        name='org-test-1',
        followers=0
    ))
    search_client.index_organization(OrganizationFactory(
        name='org-test-2',
        followers=3
    ))
    search_client.index_reuse(ReuseFactory(
        title='reuse-test-1',
        followers=0
    ))
    search_client.index_reuse(ReuseFactory(
        title='reuse-test-2',
        followers=3
    ))
    search_client.index_dataservice(DataserviceFactory(
        title='dataservice-test-1',
        followers=0
    ))
    search_client.index_dataservice(DataserviceFactory(
        title='dataservice-test-2',
        followers=3
    ))
    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    # Sort ascending
    results_number, res = search_client.query_datasets(None, 0, 20, {}, sort='followers')
    assert res[0]['title'] == 'data-test-1'
    results_number, res = search_client.query_organizations(None, 0, 20, {}, sort='followers')
    assert res[0]['name'] == 'org-test-1'
    results_number, res = search_client.query_reuses(None, 0, 20, {}, sort='followers')
    assert res[0]['title'] == 'reuse-test-1'
    results_number, res = search_client.query_dataservices(None, 0, 20, {}, sort='followers')
    assert res[0]['title'] == 'dataservice-test-1'

    # Sort descending
    results_number, res = search_client.query_datasets(None, 0, 20, {}, sort='-followers')
    assert res[0]['title'] == 'data-test-2'
    results_number, res = search_client.query_organizations(None, 0, 20, {}, sort='-followers')
    assert res[0]['name'] == 'org-test-2'
    results_number, res = search_client.query_reuses(None, 0, 20, {}, sort='-followers')
    assert res[0]['title'] == 'reuse-test-2'
    results_number, res = search_client.query_dataservices(None, 0, 20, {}, sort='-followers')
    assert res[0]['title'] == 'dataservice-test-2'

def test_search_reuse_with_score_functions(app, client, search_client, faker):
    # Create generic reuses without bonus/malus
    for i in range(4):
        search_client.index_reuse(ReuseFactory(
            title=f"reuse {i}",
            featured=1,
            followers=1,
            orga_followers=1,
            archived=None,
            views=9,
        ))
    # Create specific reuse with script score malus field (archived)
    search_client.index_reuse(ReuseFactory(
        title="archived reuse",
        featured=1,
        followers=1,
        orga_followers=1,
        archived=datetime.datetime.utcnow(),
        views=10
    ))
    # Create specific reuse with field value factor bonus field (views)
    search_client.index_reuse(ReuseFactory(
        title="reuse with a lot of views",
        featured=1,
        followers=1,
        orga_followers=1,
        archived=None,
        views=100000
    ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_reuses(None, 0, 20, {})
    assert results_number == 6
    assert res[0]["title"] == "reuse with a lot of views"
    assert res[-1]["title"] == "archived reuse"

def test_general_search_with_sorting_last_update(app, client, search_client, faker):
    search_client.index_dataset(DatasetFactory(
        title='data-test-1',
        last_update=datetime.date(2021, 12, 2)
    ))
    search_client.index_dataset(DatasetFactory(
        title='data-test-2',
        last_update=datetime.date(2022, 12, 2)
    ))
    search_client.index_dataset(DatasetFactory(
        title='data-test-3',
        last_update=datetime.date(2023, 12, 2)
    ))
    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(None, 0, 20, {}, sort='-last_update')
    assert res[0]['title'] == 'data-test-3'
    results_number, res = search_client.query_datasets(None, 0, 20, {}, sort='last_update')
    assert res[0]['title'] == 'data-test-1'


def test_search_dataset_with_synonym(app, client, search_client, faker):
    search_client.index_dataset(DatasetFactory(
        title='rp',
    ))

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets('recensement population', 0, 20, {})
    assert results_number == 1


def test_search_dataset_by_id(app, client, search_client, faker):
    dataset = DatasetFactory()
    search_client.index_dataset(dataset)

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(str(dataset.id), 0, 20, {})
    assert results_number == 1


def test_search_dataset_by_resource_title_and_id(app, client, search_client, faker):
    dataset = DatasetFactory()
    search_client.index_dataset(dataset)

    for i in range(4):
        search_client.index_dataset(DatasetFactory())

    # Without this, ElasticSearch does not seem to have the time to index.
    time.sleep(2)

    results_number, res = search_client.query_datasets(str(dataset.resources_ids[0]), 0, 20, {})
    assert results_number == 1

    results_number, res = search_client.query_datasets(str(dataset.resources_titles[0]), 0, 20, {})
    assert results_number == 1




from abc import ABC, abstractmethod
from typing import Tuple, Optional, List
from udata_search_service.domain.entities import Dataset, Organization, Reuse, Dataservice


class SearchClient(ABC):

    @abstractmethod
    def init_indices(self) -> None:
        pass

    @abstractmethod
    def clean_indices(self) -> None:
        pass

    @abstractmethod
    def index_organization(self, to_index: Organization) -> None:
        pass

    @abstractmethod
    def index_dataset(self, to_index: Dataset) -> None:
        pass

    @abstractmethod
    def index_reuse(self, to_index: Reuse) -> None:
        pass

    @abstractmethod
    def index_dataservice(self, to_index: Dataservice) -> None:
        pass

    @abstractmethod
    def query_organizations(self, query_text: str, offset: int, page_size: int) -> Tuple[int, List[Organization]]:
        pass

    @abstractmethod
    def query_datasets(self, query_text: str, offset: int, page_size: int) -> Tuple[int, List[Dataset]]:
        pass

    @abstractmethod
    def query_reuses(self, query_text: str, offset: int, page_size: int) -> Tuple[int, List[Reuse]]:
        pass

    @abstractmethod
    def query_dataservices(self, query_text: str, offset: int, page_size: int) -> Tuple[int, List[Dataservice]]:
        pass

    @abstractmethod
    def find_one_organization(self, organization_id: str) -> Optional[Organization]:
        pass

    @abstractmethod
    def find_one_dataset(self, dataset_id: str) -> Optional[Dataset]:
        pass

    @abstractmethod
    def find_one_reuse(self, reuse_id: str) -> Optional[Reuse]:
        pass

    @abstractmethod
    def find_one_dataservice(self, dataservice_id: str) -> Optional[Dataservice]:
        pass



import datetime
import factory

from udata_search_service.domain.entities import Dataservice, Dataset, Organization, Reuse


class DatasetFactory(factory.Factory):
    class Meta:
        model = Dataset

    id = factory.Faker('md5')
    title = factory.Faker('sentence')
    description = factory.Faker('text')
    acronym = factory.Faker('company_suffix')
    url = factory.Faker('url')
    created_at = factory.LazyFunction(datetime.datetime.utcnow)
    last_update = factory.LazyFunction(datetime.datetime.utcnow)
    orga_sp = 4
    orga_followers = factory.Faker('random_int')
    views = factory.Faker('random_int')
    followers = factory.Faker('random_int')
    reuses = factory.Faker('random_int')
    featured = factory.Faker('random_int')
    resources_count = factory.Faker('random_int', min=1, max=15)
    resources_ids = factory.List([factory.Faker('md5') for _ in range(5)])
    resources_titles = factory.List([factory.Faker('sentence') for _ in range(5)])
    organization = factory.Faker('md5')
    organization_name = factory.Faker('company')
    organization_badges = []
    format = ['pdf']
    frequency = 'unknown'
    concat_title_org = factory.LazyAttribute(lambda obj: f'{obj.title} {obj.acronym} {obj.organization_name}')
    badges = []
    tags = []
    topics = []
    license = factory.Faker('word')
    temporal_coverage_start = factory.Faker('past_datetime')
    temporal_coverage_end = factory.Faker('past_datetime')
    granularity = factory.Faker('word')
    geozones = factory.Faker('word')
    owner = factory.Faker('md5')


class OrganizationFactory(factory.Factory):
    class Meta:
        model = Organization

    id = factory.Faker('md5')
    name = factory.Faker('company')
    description = factory.Faker('text')
    url = factory.Faker('url')
    orga_sp = 4
    created_at = factory.LazyFunction(datetime.datetime.utcnow)
    followers = factory.Faker('random_int')
    datasets = factory.Faker('random_int')
    views = factory.Faker('random_int')
    reuses = factory.Faker('random_int')


class ReuseFactory(factory.Factory):
    class Meta:
        model = Reuse

    id = factory.Faker('md5')
    title = factory.Faker('sentence')
    description = factory.Faker('text')
    url = factory.Faker('url')
    created_at = factory.LazyFunction(datetime.datetime.utcnow)
    archived = factory.LazyFunction(datetime.datetime.utcnow)
    orga_followers = factory.Faker('random_int')
    views = factory.Faker('random_int')
    followers = factory.Faker('random_int')
    datasets = factory.Faker('random_int')
    featured = factory.Faker('random_int')
    organization = factory.Faker('md5')
    organization_name = factory.Faker('company')
    organization_badges = []
    type = factory.Faker('word')
    topic = factory.Faker('word')
    owner = factory.Faker('md5')


class DataserviceFactory(factory.Factory):
    class Meta:
        model = Dataservice

    id = factory.Faker('md5')
    title = factory.Faker('sentence')
    description = factory.Faker('text')
    description_length = factory.Faker("random_int")
    created_at = factory.LazyFunction(datetime.datetime.utcnow)
    views = factory.Faker('random_int')
    followers = factory.Faker("random_int")
    is_restricted = factory.Faker("boolean")
    organization = factory.Faker('md5')
    organization_name = factory.Faker('company')
    owner = factory.Faker('md5')



import logging
import os
import sys
from typing import List

from udata_search_service.infrastructure.search_clients import ElasticClient
from udata_search_service.config import Config

ELASTIC_HOST = os.environ.get('ELASTIC_HOST', 'localhost')
ELASTIC_PORT = os.environ.get('ELASTIC_PORT', '9200')

ALL_INDICES = [
    'dataset',
    'reuse',
    'organization',
    'dataservice',
]


def set_alias(search_client: ElasticClient, index_suffix_name: str, indices: List[str] = None, delete: bool = True) -> None:
    '''
    Properly end an indexation by swapping alias.
    Previous alias is deleted if needed.
    Alias will be removed from previous indices and set to
    the new indices (that match the suffix `index_suffix_name`).
    A list of indices on which to apply the alias migration can be passed.
    See an example logic https://github.com/elastic/elasticsearch-dsl-py/blob/master/examples/alias_migration.py
    '''

    indices = indices or ALL_INDICES
    for index in indices:
        if index not in ALL_INDICES:
            logging.error('Unknown index alias %s', index)
            sys.exit(-1)

        index_alias = Config.UDATA_INSTANCE_NAME + '-' + index
        pattern = index_alias + '-*'
        index_name = index_alias + '-' + index_suffix_name
        logging.info('Creating alias "%s" on index "%s"', index_alias, index_name)

        previous_indices = None
        if delete:
            if search_client.es.indices.exists_alias(name=index_alias):
                previous_indices = search_client.es.indices.get_alias(name=index_alias).keys()

        search_client.es.indices.update_aliases(
            body={
                "actions": [
                    {"remove": {"alias": index_alias, "index": pattern}},
                    {"add": {"alias": index_alias, "index": index_name}},
                ]
            }
        )

        if delete and previous_indices:
            for previous_index in previous_indices:
                if previous_index != index_name:
                    search_client.es.indices.delete(index=previous_index)



import click
from flask import Flask
from flask.cli import FlaskGroup
from udata_search_service.config import Config
from udata_search_service.container import Container
from udata_search_service.presentation import api, commands


def create_app(config: object = Config) -> Flask:
    container = Container()
    container.wire(modules=[api, commands])

    app: Flask = Flask(__name__)
    app.container = container
    app.config.from_object(config)

    container.config.elasticsearch_url.from_value(app.config['ELASTICSEARCH_URL'])
    container.config.search_synonyms.from_value(app.config['SEARCH_SYNONYMS'])

    app.register_blueprint(api.bp)

    return app


@click.group(cls=FlaskGroup, create_app=create_app)
def cli():
    '''udata-search-service management client'''


commands.init_app(cli)



import dataclasses
from typing import List
from datetime import datetime
from dateutil.parser import isoparse


@dataclasses.dataclass
class EntityBase():

    @classmethod
    def load_from_dict(cls, data):
        fields = [f.name for f in dataclasses.fields(cls)]
        data = {key: data[key] for key in data if key in fields}
        return cls(**data)

    def to_dict(self):
        return dataclasses.asdict(self)


@dataclasses.dataclass
class Organization(EntityBase):
    id: str
    name: str
    description: str
    url: str
    orga_sp: int
    created_at: datetime.date
    followers: int
    datasets: int
    views: int
    reuses: int

    badges: List[str] = None
    acronym: str = None

    def __post_init__(self):
        if isinstance(self.created_at, str):
            self.created_at = isoparse(self.created_at)


@dataclasses.dataclass
class Dataset(EntityBase):
    id: str
    title: str
    url: str
    created_at: datetime.date
    frequency: str
    format: List[str]
    views: int
    followers: int
    reuses: int
    featured: int
    resources_count: int
    concat_title_org: str
    description: str

    last_update: datetime.date = None
    acronym: str = None
    badges: List[str] = None
    tags: List[str] = None
    license: str = None
    temporal_coverage_start: datetime.date = None
    temporal_coverage_end: datetime.date = None
    granularity: str = None
    geozones: List[str] = None
    schema: List[str] = None
    topics: List[str] = None
    resources_ids: List[str] = None
    resources_titles: List[str] = None

    orga_sp: int = None
    orga_followers: int = None
    organization: str = None
    organization_name: str = None
    organization_badges: List[str] = None
    owner: str = None

    def __post_init__(self):
        if isinstance(self.created_at, str):
            self.created_at = isoparse(self.created_at)
        if isinstance(self.last_update, str):
            self.last_update = isoparse(self.last_update)
        if isinstance(self.temporal_coverage_start, str):
            self.temporal_coverage_start = isoparse(self.temporal_coverage_start)
        if isinstance(self.temporal_coverage_end, str):
            self.temporal_coverage_end = isoparse(self.temporal_coverage_end)


@dataclasses.dataclass
class Reuse(EntityBase):
    id: str
    title: str
    url: str
    created_at: datetime.date
    views: int
    followers: int
    datasets: int
    featured: int
    description: str
    type: str
    topic: str

    archived: datetime.date = None
    tags: List[str] = None
    badges: List[str] = None
    orga_followers: int = None
    organization: str = None
    organization_name: str = None
    organization_badges: List[str] = None
    owner: str = None

    def __post_init__(self):
        if isinstance(self.created_at, str):
            self.created_at = isoparse(self.created_at)
        if isinstance(self.archived, str):
            self.archived = isoparse(self.archived)

@dataclasses.dataclass
class Dataservice(EntityBase):
    id: str
    title: str
    description: str
    description_length: float
    created_at: datetime.date

    views: int = 0
    followers: int = 0
    is_restricted: bool = None
    orga_followers: int = None
    organization: str = None
    organization_name: str = None
    owner: str = None
    tags: List[str] = None

    def __post_init__(self):
        if isinstance(self.created_at, str):
            self.created_at = isoparse(self.created_at)



'''
udata search service
'''

import os

tag = os.environ.get('CIRCLE_TAG')
build_num = os.environ.get('CIRCLE_BUILD_NUM')

__version__ = '2.2.3.dev' + (str(build_num) if not tag and build_num else '')



from dependency_injector import containers, providers
from udata_search_service.infrastructure.services import DatasetService, OrganizationService, ReuseService, DataserviceService
from udata_search_service.infrastructure.search_clients import ElasticClient


class Container(containers.DeclarativeContainer):
    config = providers.Configuration()

    search_client = providers.Singleton(
        ElasticClient,
        url=config.elasticsearch_url
    )

    organization_service = providers.Factory(
        OrganizationService,
        search_client=search_client
    )

    dataset_service = providers.Factory(
        DatasetService,
        search_client=search_client
    )

    reuse_service = providers.Factory(
        ReuseService,
        search_client=search_client
    )

    dataservice_service = providers.Factory(
        DataserviceService,
        search_client=search_client
    )
